## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
# make_VLM_wings.py

# Created:  Jun 2021, A. Blaufox

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports 
import numpy as np
from copy import deepcopy
import  RCAIDE

from RCAIDE.Framework.Core import  Data
from RCAIDE.Library.Components.Wings import All_Moving_Surface 
from RCAIDE.Library.Components.Wings.Control_Surfaces import Aileron , Elevator , Slat , Flap , Rudder 
from RCAIDE.Library.Methods.Geometry.Planform import populate_control_sections
from RCAIDE.Library.Methods.Geometry.Planform.convert_sweep import convert_sweep_segments

# ------------------------------------------------------------------
# make_VLM_wings()
# ------------------------------------------------------------------  
## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def make_VLM_wings(geometry, settings):
    """ This parses through geometry.wings to create a Container of Data objects.
        Relevant VLM attributes are copied from geometry.wings to the Container.
        After, the wing data objects are reformatted. All control surfaces are 
        also added to the Container as Data objects representing full wings. 
        Helper variables are then computed (most notably span_breaks) for later. 
        
        see make_span_break() for further details

    Assumptions: 
    All control surfaces are appended directly to the wing, not wing segments.
    If a given wing has no segments, it must have either .taper or .chords.root 
        and .chords.tip defined

    Source:   
    None
    
    Inputs:
    geometry.
        wings.wing.
            twists.root
            twists.tip
            dihedral
            sweeps.quarter_chord OR sweeps.leading_edge
            thickness_to_chord
            taper
            chord.root
            chords.tip
            
            control_surface.
                tag
                span_fraction_start
                span_fraction_end
                deflection
                chord_fraction
                
    settings.discretize_control_surfaces  --> set to True to generate control surface panels
    
    Properties Used:
    N/A
    """ 
    # unpack inputs
    discretize_cs = settings.discretize_control_surfaces
    wings         = copy_wings(geometry.wings)
    
    # ------------------------------------------------------------------
    # Reformat original wings to have at least 2 segments and additional values for processing later
    # ------------------------------------------------------------------    
    for wing in wings:
        wing.is_a_control_surface = False
        n_segments           = len(wing.segments.keys())
        if n_segments==0:
            # convert to preferred format for the panelization loop
            wing       = convert_to_segmented_wing(wing)
            n_segments = 2
        else:
            # check for invalid/unsupported/conflicting geometry input            
            if issubclass(wing.wing_type, All_Moving_Surface): # these cases unsupported due to the way the panelization loop is structured at the moment
                if not (wing.hinge_vector == np.array([0.,0.,0.])).all() and wing.use_constant_hinge_fraction:
                    raise ValueError("A hinge_vector is specified, but the surface is set to use a constant hinge fraction")
                if len(wing.control_surfaces) > 0:
                    raise ValueError('Input: control surfaces are not supported on all-moving surfaces at this time')
            for segment in wing.segments: #unsupported by convention
                if 'control_surfaces' in segment.keys() and len(segment.control_surfaces) > 0:
                    raise ValueError('Input: control surfaces should be appended to the wing, not its segments. ' + 
                                     'This function will move the control surfaces to wing segments itself.')  
        
        #move wing control surfaces to from wing to its segments
        wing = populate_control_sections(wing) if discretize_cs else wing
        
        #ensure wing has attributes that will be needed later
        wing_halfspan = wing.spans.projected * 0.5 if wing.symmetric else wing.spans.projected
        segment_list  = list(wing.segments.keys())
        for i in range(n_segments):   
            (ia, ib)       = (0, 0) if i==0 else (i-1, i)
            seg_a          = wing.segments[segment_list[ia]]
            seg_b          = wing.segments[segment_list[ib]]            
            seg_b.chord    = seg_b.root_chord_percent *wing.chords.root  ##may be worth implementing a self-calculating .chord attribute    
            
            #guarantee that all segments have leading edge sweep
            if (i != 0) and (seg_a.sweeps.leading_edge is None):
                old_sweep                 = seg_a.sweeps.quarter_chord
                new_sweep                 = convert_sweep_segments(old_sweep, seg_a, seg_b, wing, old_ref_chord_fraction=0.25, new_ref_chord_fraction=0.0)
                seg_a.sweeps.leading_edge = new_sweep 
                
            #give segments offsets for giving cs_wings an origin later
            section_span     = (seg_b.percent_span_location - seg_a.percent_span_location) * wing_halfspan
            seg_b.x_offset   = 0. if i==0 else seg_a.x_offset   + section_span*np.tan(seg_a.sweeps.leading_edge)
            seg_b.dih_offset = 0. if i==0 else seg_a.dih_offset + section_span*np.tan(seg_a.dihedral_outboard)
        wing.segments[segment_list[-1]].sweeps.leading_edge = 1e-8
    
    # each control_surface-turned-wing will have its own unique ID number
    cs_ID = 0
    
    # ------------------------------------------------------------------
    # Build wing Data() objects and wing.span_breaks from control surfaces on segments
    # ------------------------------------------------------------------    
    for wing in wings:
        if wing.is_a_control_surface == True: #skip if this wing is actually a control surface
            continue
        
        #prepare to iterate across all segments and control surfaces
        seg_breaks  = RCAIDE.Framework.Core.ContainerOrdered()
        LE_breaks   = RCAIDE.Framework.Core.ContainerOrdered()
        TE_breaks   = RCAIDE.Framework.Core.ContainerOrdered()
        n_segments  = len(wing.segments.keys())

        #process all control surfaces in each segment-------------------------------------
        segment_list  = list(wing.segments.keys())
        for i in range(n_segments):   
            (ia, ib)    = (0, 0) if i==0 else (i-1, i)
            seg_a       = wing.segments[segment_list[ia]]
            seg_b       = wing.segments[segment_list[ib]]            
            
            control_surfaces = seg_b.control_surfaces if 'control_surfaces' in seg_b.keys() else Data()
            for cs in control_surfaces: #should be no control surfaces on root segment
                # create and append a wing object from the control_surface object and relevant segments
                cs_wing = make_cs_wing_from_cs(cs, seg_a, seg_b, wing, cs_ID)
                wings.append(cs_wing)
                                
                # register cs start and end span breaks
                cs_span_breaks = make_span_breaks_from_cs(cs, seg_a, seg_b, cs_wing, cs_ID)
                if cs.cs_type==Slat:
                    LE_breaks.append(cs_span_breaks[0])
                    LE_breaks.append(cs_span_breaks[1])
                else:
                    TE_breaks.append(cs_span_breaks[0])
                    TE_breaks.append(cs_span_breaks[1])                    
                cs_ID += 1
            
            # register segment span break
            span_break = make_span_break_from_segment(seg_b)
            seg_breaks.append(span_break)

        #merge _breaks arrays into one span_breaks array----------------------------------
        #   1.  sort all span_breaks by their span_fraction
        #   2.  combine LE and TE breaks with the same span_fraction values (LE cuts from slats and TE cuts from others)
        #   3.  scan LE and TE to pick up cs cuts that cross over one or more span breaks
        
        # 1: 
        LE_breaks  = sorted(LE_breaks,  key=lambda span_break: span_break.span_fraction)
        TE_breaks  = sorted(TE_breaks,  key=lambda span_break: span_break.span_fraction)
        seg_breaks = sorted(seg_breaks, key=lambda span_break: span_break.span_fraction)
        
        # 2: similar to a 3-way merge sort
        span_breaks = RCAIDE.Framework.Core.ContainerOrdered()        
        n_LE        = len(LE_breaks)
        n_TE        = len(TE_breaks)
        n_seg       = len(seg_breaks)
        i, j, k     = 0,0,0
        big_num     = float('inf')
        while True:
            LE_span  = LE_breaks[i].span_fraction  if (i < n_LE)  else big_num
            TE_span  = TE_breaks[j].span_fraction  if (j < n_TE)  else big_num
            seg_span = seg_breaks[k].span_fraction if (k < n_seg) else big_num
            
            if (LE_span==big_num) and (TE_span==big_num) and (seg_span==big_num):
                break
            
            if (LE_span <= TE_span) and (LE_span <= seg_span):
                add_span_break(LE_breaks[i], span_breaks)
                i += 1
            elif (TE_span <= LE_span) and (TE_span <= seg_span):
                add_span_break(TE_breaks[j], span_breaks)
                j += 1  
            elif (seg_span <= LE_span) and (seg_span <= TE_span):
                add_span_break(seg_breaks[k], span_breaks)
                k += 1   
            else:
                raise ValueError("No suitable span break") #should never occur
                
        # 3:
        ib, ob = 0, 1 #inboard, outboard indices
        for edge, edge_str in enumerate(['LE','TE']):
            for i in range(len(span_breaks)-1):
                ID_i = span_breaks[i].cs_IDs[edge,ob]
                cut  = span_breaks[i].cuts[edge,ob]
                if ID_i == -1:
                    continue
                #copy the cs ID and its cut until the end of the control surface is found
                for j in range(i+1,len(span_breaks)):
                    i    += 1
                    ID_j = span_breaks[j].cs_IDs[edge,ib]                    
                    if ID_j == ID_i: #found control surface end
                        break
                    elif ID_j == -1: #found a span_break within control surface. copy values
                        span_breaks[j].cs_IDs[edge,:] = [ID_i, ID_i]
                        span_breaks[j].cuts[edge,:]   = [cut, cut]
                    else:
                        raise ValueError('VLM does not support multiple control surfaces on the same edge at this time')
                
        # pack span_breaks
        wing.span_breaks = reprocess_span_breaks(span_breaks)
        
    # ------------------------------------------------------------------
    # Give cs_wings span_breaks arrays
    # ------------------------------------------------------------------   
    for cs_wing in wings:
        cs_w_segs = list(cs_wing.segments.keys())
        if cs_wing.is_a_control_surface == False: #skip if this wing isn't actually a control surface
            continue  
        span_breaks = RCAIDE.Framework.Core.ContainerOrdered()
        span_break  = make_span_break_from_segment(cs_wing.segments[cs_w_segs[0]])
        span_breaks.append(span_break)
        span_break  = make_span_break_from_segment(cs_wing.segments[cs_w_segs[1]])
        span_breaks.append(span_break) 
        cs_wing.span_breaks = span_breaks
    
    return wings
  

# ------------------------------------------------------------------
# custom deepcopy(wings)
# --TO DO-- This is a stand-in for a more fleshed-out VLM_surface class
# ------------------------------------------------------------------      
def copy_wings(original_wings):
    """ This copies VLM attributes for every wing object in original_wings into 
    a new wings container with new Data objects
    
    Inputs:   
    original_wings - the original wings container
    """       
    return copy_large_container(original_wings, "wings")

def copy_large_container(large_container, type_str):
    """ This function helps avoid copying a container of large objects directly,
    especially if those objects are Physical_Components
    
    Inputs:
    objects  -  a Container of large objects
    """    
    container = RCAIDE.Framework.Core.Container()  if type_str != "Segments" else RCAIDE.Framework.Core.ContainerOrdered()
    paths = get_paths(type_str)
    
    for obj in large_container:
        if type(obj) ==  RCAIDE.Library.Components.Wings.Control_Surfaces.Slat: # DO NOT COPY SLATS
            continue
        else:
            #copy from paths
            data = copy_data_from_paths(obj, paths)     
            
            #special case new attributes
            if type_str == 'control_surfaces':
                data.cs_type                     = type(obj) # needed to identify the class of a control surface
            elif type_str == 'wings':
                data.wing_type = type(obj)
                if issubclass(data.wing_type, All_Moving_Surface):
                    data.sign_duplicate              = obj.sign_duplicate
                    data.hinge_fraction              = obj.hinge_fraction 
                    data.deflection                  = obj.deflection  
                    data.is_slat                     = False
                    data.use_constant_hinge_fraction = obj.use_constant_hinge_fraction
                    data.hinge_vector                = obj.hinge_vector
                    data.deflection_last             = 0.
            container.append(data)
        
    return container

def copy_data_from_paths(old_object, paths):
    """ This copies the attributes specified by 'paths' from old_object  
    into a new Data() object

    Inputs:   
    old_object - an object to copy
    """       
    new_object = Data()   
    for path in paths:
        val = old_object.deep_get(path)
        recursive_set(new_object, path, val)
    return new_object

def recursive_set(data_obj, path, val):
    """ This is similar to the deep_set function, but also creates
    intermediate Data() objects for keys that do not yet exist. Special
    copy cases are made for paths that lead to large class objects
    """
    special_case_keys = ['control_surfaces', 'segments']
    keys = path.split('.')
    key  = keys[0]
    if len(keys) == 1:
        if key in special_case_keys:
            data_obj[key] = copy_large_container(val, key) # will eventually recurse back to this function
        else:
            data_obj[key] = deepcopy(val) # at this point, should only be copying primitive types or very small Data objects
        return
    
    has_key = key in data_obj.keys()
    if not has_key:
        data_obj[key] = Data()
        
    new_path = '.'.join(keys[1:])
    recursive_set(data_obj[key], new_path, val)

def get_paths(type_str):
    """ This returns a list of the paths to the attributes needed in VLM
    for a given type of object.
    
    Note that if any element in the paths array is the same as the array's correponding type_str, 
    this will cause copy_large_container() to recurse infinitely. It will also recurse infinitely 
    if any element in the current array is the same as a type_str that corresponds to a different array
    which itself has an element that is the same as the current type_str. 
    
    Inputs:
    type_str  -  "wings", "control_surfaces" or "Segments"
    """       
    if type_str == 'wings':
        paths = ['tag',
                'origin',
                'symmetric',
                'vertical',
                'taper',
                'dihedral',
                'thickness_to_chord',
                'spans.projected',
                'chords.root',
                'chords.tip',
                'sweeps.quarter_chord',
                'sweeps.leading_edge',
                'twists.root',
                'twists.tip',
                'vortex_lift',
                'airfoil',
                'segments',
                'control_surfaces',
                ]
    elif type_str == 'control_surfaces':
        paths = ['tag',               
                 'span',               
                 'span_fraction_start',
                 'span_fraction_end',  
                 'hinge_fraction',     
                 'chord_fraction',     
                 'sign_duplicate',
                 'deflection',         
                 'configuration_type',      
                 'gain',      
                 ]
    elif type_str == 'segments':
        paths = ['tag',               
                 'percent_span_location',               
                 'twist',
                 'root_chord_percent',  
                 'dihedral_outboard',     
                 'thickness_to_chord',     
                 'sweeps.quarter_chord',         
                 'sweeps.leading_edge', 
                 'airfoil', 
                 ]        
        
    return paths

# ------------------------------------------------------------------
# wing helper functions
# ------------------------------------------------------------------  
def make_cs_wing_from_cs(cs, seg_a, seg_b, wing, cs_ID):
    """ This uses a control surface and the segment it lies between to create
    an equilvalent wing object. The wing has a couple of non-standard attributes
    that contain information about the control surface it came from

    Assumptions: 
    None

    Source:   
    None
    
    Inputs:   
    cs    - a control surface object
    seg_a - the segment object inboard of the cs
    seg_b - the segment object outboard of the cs. The cs is also attached to this
    wing  - the wing object which owns seg_a and seg_b
    cs_ID - a unique identifier for the cs_wing
    
    Outputs:
    cs_wing - a Data object with relevant Wing and Control_Surface attributes
    
    Properties Used:
    N/A
    """      
    hspan = wing.spans.projected*0.5 if wing.symmetric else wing.spans.projected
    
    cs_wing                       = copy_data_from_paths(RCAIDE.Library.Components.Wings.Wing(), get_paths("wings"))
    
    #standard wing attributes--------------------------------------------------------------------------------------
    cs_wing.tag                   = wing.tag + '__cs_id_{}'.format(cs_ID)
    span_a                        = seg_a.percent_span_location
    span_b                        = seg_b.percent_span_location
    twist_a                       = seg_a.twist
    twist_b                       = seg_b.twist
    cs_wing.twists.root           = np.interp(cs.span_fraction_start, [span_a, span_b], [twist_a, twist_b])
    cs_wing.twists.tip            = np.interp(cs.span_fraction_end,   [span_a, span_b], [twist_a, twist_b])
    cs_wing.dihedral              = seg_a.dihedral_outboard
    cs_wing.thickness_to_chord    = (seg_a.thickness_to_chord + seg_b.thickness_to_chord)/2
    cs_wing.origin                = np.array(wing.origin) *1.
    
    span_fraction_tot             = cs.span_fraction_end - cs.span_fraction_start
    cs_wing.spans.projected       = wing.spans.projected * span_fraction_tot #includes 2x length if cs is on a symmetric wing 
    
    wing_chord_local_at_cs_root   = np.interp(cs.span_fraction_start, [span_a, span_b], [seg_a.chord, seg_b.chord])
    wing_chord_local_at_cs_tip    = np.interp(cs.span_fraction_end,   [span_a, span_b], [seg_a.chord, seg_b.chord])
    cs_wing.chords.root           = wing_chord_local_at_cs_root * cs.chord_fraction  
    cs_wing.chords.tip            = wing_chord_local_at_cs_tip  * cs.chord_fraction             
    cs_wing.taper                 = cs_wing.chords.tip / cs_wing.chords.root
    cs_wing.sweeps.quarter_chord  = 0.  # leave at 0. VLM will use leading edge

    cs_wing.symmetric             = wing.symmetric
    cs_wing.vertical              = wing.vertical
    cs_wing.vortex_lift           = wing.vortex_lift

    #non-standard wing attributes, mostly to do with cs_wing's identity as a control surface-----------------------
    #metadata
    cs_wing.is_a_control_surface  = True
    cs_wing.cs_ID                 = cs_ID
    cs_wing.name                  = wing.tag + '__' + seg_b.tag + '__' + cs.tag + '__cs_ID_{}'.format(cs_ID)
    cs_wing.is_slat               = (cs.cs_type==Slat)
    cs_wing.is_aileron            = (cs.cs_type==Aileron)
    cs_wing.pivot_edge            = 'TE' if cs_wing.is_slat else 'LE'
    
    #control surface attributes
    cs_wing.chord_fraction        = cs.chord_fraction
    cs_wing.hinge_fraction        = cs.hinge_fraction
    cs_wing.sign_duplicate        = cs.sign_duplicate
    cs_wing.deflection            = cs.deflection
    cs_wing.deflection_last       = 0.
    
    #adjustments---------------------------------------------------------------------------------------------------
    #adjust origin - may need to be adjusted later
    wing_halfspan                 = wing.spans.projected * 0.5 if wing.symmetric else wing.spans.projected
    LE_TE_cs_offset               = 0. if cs_wing.is_slat else (1 - cs.chord_fraction)*wing_chord_local_at_cs_root
    cs_wing.origin[0,0]          += np.interp(cs.span_fraction_start, [span_a, span_b], [seg_a.x_offset, seg_b.x_offset]) + LE_TE_cs_offset
    cs_wing.origin[0,1]          += cs.span_fraction_start * wing_halfspan if not wing.vertical else np.interp(cs.span_fraction_start, [span_a, span_b], [seg_a.dih_offset, seg_b.dih_offset])
    cs_wing.origin[0,2]          += np.interp(cs.span_fraction_start, [span_a, span_b], [seg_a.dih_offset, seg_b.dih_offset]) if not wing.vertical else cs.span_fraction_start * wing_halfspan
    
    # holds all required y-coords. Will be added to during discretization to ensure y-coords match up between wing and control surface.
    rel_offset                    = cs_wing.origin[0,1] - wing.origin[0][1] if not cs_wing.vertical else cs_wing.origin[0,2] - wing.origin[0][2]
    cs_wing.y_coords_required     = [cs.span_fraction_end*hspan - rel_offset] #initialize with the tip y-coord. Other coords to be added in VLM

    #find sweep of the 'outside' edge (LE for slats, TE for everything else)
    use_le_sweep                  = not (seg_a.sweeps.leading_edge is None)
    new_cf                        = 0. if cs_wing.is_slat else 1
    old_cf                        = 0. if use_le_sweep else 0.25
    old_sweep                     = seg_a.sweeps.leading_edge if use_le_sweep else seg_a.sweeps.quarter_chord
    new_sweep                     = convert_sweep_segments(old_sweep, seg_a, seg_b, wing, old_ref_chord_fraction=old_cf, new_ref_chord_fraction=new_cf)
    cs_wing.outside_sweep         = new_sweep
    
    #find leading edge sweep
    if cs_wing.is_slat:
        cs_wing.sweeps.leading_edge = new_sweep
    else:
        new_cf                      = 1 - cs_wing.chord_fraction
        new_sweep                   = convert_sweep_segments(old_sweep, seg_a, seg_b, wing, old_ref_chord_fraction=old_cf, new_ref_chord_fraction=new_cf)
        cs_wing.sweeps.leading_edge = new_sweep
    
    #convert to segmented wing-------------------------------------------------------------------------------------
    cs_wing = convert_to_segmented_wing(cs_wing)
    
    # give segments offsets (in coordinates relative to the cs_wing) 
    cs_w_segs = list(cs_wing.segments.keys())    
    cs_wing.segments[cs_w_segs[0]].x_offset   = 0.
    cs_wing.segments[cs_w_segs[0]].dih_offset = 0.  
    cs_wing.segments[cs_w_segs[1]].x_offset   = wing_halfspan * span_fraction_tot *np.tan(cs_wing.segments[cs_w_segs[0]].sweeps.leading_edge)
    cs_wing.segments[cs_w_segs[1]].dih_offset = wing_halfspan * span_fraction_tot *np.tan(cs_wing.segments[cs_w_segs[0]].dihedral_outboard)    
    
    #add airfoil
    cs_wing.segments[cs_w_segs[0]].airfoil     = seg_a.airfoil
    cs_wing.segments[cs_w_segs[1]].airfoil     = seg_b.airfoil if cs.span_fraction_end==span_b else seg_a.airfoil
    
    return cs_wing

def convert_to_segmented_wing(wing):
    """ This turns a non-segmented wing into a segmented wing

    Assumptions: 
    If a given wing has no segments, it must have either .taper or .chords.tip defined

    Source:   
    None
    
    Inputs:   
    VD                   - vortex distribution 
    geometry.
        wings.wing.
            twists.root
            twists.tip
            dihedral
            sweeps.quarter_chord
            thickness_to_chord
            taper
            chord.root
            chords.tip
    
    Properties Used:
    N/A
    """     
    if len(wing.segments.keys()) > 0:
        return wing   
    # root segment 
    segment                               = RCAIDE.Library.Components.Wings.Segments.Segment()
    segment.tag                           = 'root_segment'
    segment.percent_span_location         = 0.0
    segment.twist                         = wing.twists.root
    segment.root_chord_percent            = 1.
    segment.chord                         = wing.chords.root #non-standard attribute, needed for VLM
    segment.dihedral_outboard             = wing.dihedral
    segment.sweeps.quarter_chord          = wing.sweeps.quarter_chord
    segment.sweeps.leading_edge           = wing.sweeps.leading_edge
    segment.thickness_to_chord            = wing.thickness_to_chord
    if wing.airfoil: 
        segment.append_airfoil(wing.airfoil)              
    wing.segments.append(segment) 
    
    # tip segment 
    if wing.taper==0:
        wing.taper = wing.chords.tip / wing.chords.root
    elif wing.chords.tip==0:
        wing.chords.tip = wing.chords.root * wing.taper
        
    segment                               = RCAIDE.Library.Components.Wings.Segments.Segment()
    segment.tag                           = 'tip_segment'
    segment.percent_span_location         = 1.
    segment.twist                         = wing.twists.tip
    segment.root_chord_percent            = wing.taper
    segment.chord                         = wing.chords.tip #non-standard attribute, needed for VLM
    segment.dihedral_outboard             = 0.
    segment.sweeps.quarter_chord          = 0.
    segment.sweeps.leading_edge           = 1e-8
    segment.thickness_to_chord            = wing.thickness_to_chord
    if wing.airfoil: 
        segment.append_airfoil(wing.airfoil)             
    wing.segments.append(segment) 
    
    return wing

# ------------------------------------------------------------------
# span_break processing helper functions
# ------------------------------------------------------------------  
def add_span_break(span_break, span_breaks):
    """ This is a helper function that appends or superimposes a span_break 
    into span_breaks

    Assumptions: 
    None

    Source:   
    None
    
    Inputs:   
    span_break
    span_breaks
    
    Properties Used:
    N/A
    """     
    if len(span_breaks) == 0:
        span_breaks.append(span_break)
    else:
        # if non-coincident, the space between the breaks is nominal wing: append the new span_break
        if span_breaks[-1].span_fraction < span_break.span_fraction: 
            span_breaks.append(span_break)
        
        # else coincident: need to superimpose cs_IDs and cuts, not append
        else:
            boolean = span_breaks[-1].cs_IDs==-1
            span_breaks[-1].cs_IDs[boolean] = span_break.cs_IDs[boolean]
            span_breaks[-1].cuts[boolean]   = span_break.cuts[boolean]
                
    return


def reprocess_span_breaks(span_breaks):
    """ This reprocesses the tags in a newly superimposed set of
    span_breaks and creates a new object so that the new keys match 
    the new tags
    
    Inputs:
    span_breaks
    """     
    sbs = RCAIDE.Framework.Core.ContainerOrdered()
    for i,span_break in enumerate(span_breaks):
        span_break.tag = make_span_break_tag(span_break)
        sbs.append(span_break)
    return sbs

# ------------------------------------------------------------------
# span_break creation helper functions
# ------------------------------------------------------------------ 
def make_span_break_from_segment(seg):
    """ This creates a span_break Data() object from a segment

    Assumptions: 
    None

    Source:   
    None
    
    Inputs:   
    seg    - a segment object with standard attributes except for:
        .chord
    
    Properties Used:
    N/A
    """       
    span_frac       = seg.percent_span_location
    airfoil         = seg.airfoil
    dihedral_ob     = seg.dihedral_outboard
    sweep_ob_QC     = seg.sweeps.quarter_chord
    sweep_ob_LE     = seg.sweeps.leading_edge
    twist           = seg.twist    
    local_chord     = seg.chord       #non-standard attribute
    x_offset        = seg.x_offset
    dih_offset      = seg.dih_offset   
    span_break = make_span_break(-1, 0, 0, span_frac, 0., airfoil,
                                 dihedral_ob, sweep_ob_QC, sweep_ob_LE, twist, local_chord,
                                 x_offset, dih_offset)  
    span_break.cuts = np.array([[0.,0.],  
                                [1.,1.]])
    return span_break

def make_span_breaks_from_cs(cs, seg_a, seg_b, cs_wing, cs_ID):
    """ This creates span_break Data() objects from a control surface, its
    owning segments, and their owning cs_wing

    Assumptions: 
    None

    Source:   
    None
    
    Inputs:   
    cs      - a control surface object
    seg_a   - the segment object inboard of the cs
    seg_b   - the segment object outboard of the cs. The cs is also attached to this
    cs_wing - the wing object which owns seg_a and seg_b
    cs_ID   - a unique identifier for the cs_wing
    
    Properties Used:
    N/A
    """     
    is_slat        = (cs.cs_type==Slat)
    LE_TE          = 0 if is_slat else 1
    span_a         = seg_a.percent_span_location
    span_b         = seg_b.percent_span_location    
    
    #inboard span break
    ib_ob          = 1 #the inboard break of the cs is the outboard part of the span_break
    span_frac      = cs.span_fraction_start    
    ob_cut         = cs.chord_fraction if is_slat else 1 - cs.chord_fraction
    airfoil        = seg_a.airfoil
    dihedral_ob    = seg_a.dihedral_outboard
    sweep_ob_QC    = seg_a.sweeps.quarter_chord
    sweep_ob_LE    = seg_a.sweeps.leading_edge
    twist          = cs_wing.twists.root    
    local_chord    = cs_wing.chords.root / cs.chord_fraction
    x_offset       = np.interp(cs.span_fraction_start, [span_a, span_b], [seg_a.x_offset, seg_b.x_offset])
    dih_offset     = np.interp(cs.span_fraction_start, [span_a, span_b], [seg_a.dih_offset, seg_b.dih_offset])
    inboard_span_break  = make_span_break(cs_ID, LE_TE, ib_ob, span_frac, ob_cut, airfoil,
                                          dihedral_ob, sweep_ob_QC, sweep_ob_LE, twist, local_chord,
                                          x_offset, dih_offset)
    
    #outboard span break
    is_coincident  = (cs.span_fraction_end==seg_b.percent_span_location)
    ib_ob          = 0 #the outboard break of the cs is the inboard part of the span_break
    span_frac      = cs.span_fraction_end
    ib_cut         = cs.chord_fraction if is_slat else 1 - cs.chord_fraction
    airfoil        = seg_b.airfoil               if is_coincident else seg_a.airfoil #take seg_b value if this outboard break is conicident with seg_b 
    dihedral_ob    = seg_b.dihedral_outboard     if is_coincident else seg_a.dihedral_outboard
    sweep_ob_QC    = seg_b.sweeps.quarter_chord  if is_coincident else seg_a.sweeps.quarter_chord
    sweep_ob_LE    = seg_b.sweeps.leading_edge   if is_coincident else seg_a.sweeps.leading_edge
    twist          = cs_wing.twists.tip    
    local_chord    = cs_wing.chords.tip  / cs.chord_fraction
    x_offset       = np.interp(cs.span_fraction_end, [span_a, span_b], [seg_a.x_offset, seg_b.x_offset])
    dih_offset     = np.interp(cs.span_fraction_end, [span_a, span_b], [seg_a.dih_offset, seg_b.dih_offset])    
    outboard_span_break = make_span_break(cs_ID, LE_TE, ib_ob, span_frac, ib_cut, airfoil,
                                          dihedral_ob, sweep_ob_QC, sweep_ob_LE, twist, local_chord,
                                          x_offset, dih_offset)    
    return inboard_span_break, outboard_span_break

def make_span_break(cs_ID, LE_TE, ib_ob, span_frac, chord_cut, airfoil,
                    dihedral_ob, sweep_ob_QC, sweep_ob_LE, twist, local_chord,
                    x_offset, dih_offset):
    """ This gathers information related to a span break into one Data() object.
    A span break is the spanwise location of a discontinuity in the discretization
    of the panels. These can be caused by segments and by the inboard and outboard 
    edges of a control surface. The inboard and outboard sides of a span break can
    have different chords due to cuts made by control surfaces. Ultimately, the
    attributes of the span_breaks of the wing will provide the discretization function 
    generate_wing_vortex_distribution() with the necessary values to make VLM panels
    as well as reshape those panels to make the control surface cuts dipicted below.
    
    A diagram is given below:


                                            nominal local chord
    fuselage                inboard LE  |           |           .  outboard LE
    <---                                |           |           . 
                                        |           |           .
                                        |           |           .   <-- cut from a slat
                                        |           |           |
                                        |           |           |
                                        |           |           |
                                        |           |           |
                                        |           |           |
                                        |           |           |
                                        |           |           |
                                        |           |           .   <-- cut from a non-slat control surface with a different chord 
cut from a non-slat control surface     |           |           .       fraction than the control surface on the inboard side
                                  -->   .           |           .
                                        .           |           .
                            inboard TE  .           |           .  outboard TE
                                        
                                        
                                        
                                        |_______________________|
                                                    |
                                            there is 0 spanwise 
                                            distance between inboard 
                                            and outboard sides

    Outputs:
    span_break
    
    Properties Used:
    N/A
    """     
    span_break = Data()
    span_break.cs_IDs               = np.array([[-1,-1],  #  [[inboard LE cs, outboard LE cs],
                                                [-1,-1]]) #   [inboard TE cs, outboard TE cs]]
    span_break.cs_IDs[LE_TE,ib_ob]  = cs_ID
    span_break.span_fraction        = span_frac
    # The following 'cut' attributes are in terms of the local total chord and represent positions. 
    #    (an aileron with chord fraction 0.2 would have a cut value of 0.8)
    # For inboard_cut, -1 takes value of previous outboard cut value in a later function
    # For outboard_cut, -1 takes value of next inboard cut value.
    # If no break directly touching this one, cut becomes 0 (LE) or 1 (TE).
    span_break.cuts                 = np.array([[0.,0.],   #  [[inboard LE cut, outboard LE cut],
                                                [1.,1.]])  #   [inboard TE cut, outboard TE cut]]
    span_break.cuts[LE_TE,ib_ob]    = chord_cut
    span_break.airfoil              = airfoil
    span_break.dihedral_outboard    = dihedral_ob
    span_break.sweep_outboard_QC    = sweep_ob_QC
    span_break.sweep_outboard_LE    = sweep_ob_LE
    span_break.twist                = twist
    span_break.local_chord          = local_chord #this is the local chord BEFORE cuts are made
    span_break.x_offset             = x_offset
    span_break.dih_offset           = dih_offset  #dih_offset is the y or z accumulated offset from dihedral
    span_break.tag                  = make_span_break_tag(span_break)
    return span_break

def make_span_break_tag(span_break):
    location   = round(span_break.span_fraction, 3)
    cs_IDs_arr = span_break.cs_IDs.flatten()
    cs_IDs_str = '{}'.format(cs_IDs_arr).replace('[','').replace(']','').replace('-1', 'na').replace('  ', '_')
    
    return "{}___{}".format(location, cs_IDs_str)



# RCAIDE/Methods/Energy/Propulsors/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Common
from . import Constant_Speed_ICE_Propulsor
from . import Converters 
from . import ICE_Propulsor
from . import Electric_Rotor_Propulsor
from . import Electric_Ducted_Fan_Propulsor
from . import Modulators
from . import Turbofan_Propulsor
from . import Turbojet_Propulsor
from . import Turboshaft_Propulsor


# RCAIDE/Methods/Performance/estimate_stall_speed.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
import RCAIDE
 
# Pacakge imports 
import numpy as np  

#------------------------------------------------------------------------------
# Stall Speed Estimation
#------------------------------------------------------------------------------ 
def estimate_stall_speed(vehicle_mass,reference_area,altitude,maximum_lift_coefficient): 
    """Calculates the stall speed of an aircraft at a given altitude and a maximum lift coefficient.

        Sources:
        N/A

        Assumptions:
        None 

        Inputs:
            vehicle_mass                    vehicle mass             [kg]
            reference_area                  vehicle reference area   [m^2] 
            altitude                        cruise altitude          [m]
            maximum_lift_coefficient        maximum lift coefficient [unitless] 
            
        Outputs: 
            V_stall                         stall speed              [m/s]
    """ 
      
    g       = 9.81 
    atmo    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    rho     = atmo.compute_values(altitude,0.).density 
    V_stall = float(np.sqrt((2.*vehicle_mass*g)/(rho*reference_area*maximum_lift_coefficient)))  
    
    return V_stall


# RCAIDE/Methods/Energy/Propulsors/Turbofan_Propulsor/compute_turbofan_performance.py
# 
# 
# Created:  Jul 2024, RCAIDE Team

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Core import Data   
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Fan                import compute_fan_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turbofan_Propulsor            import compute_thrust

import  numpy as  np
from copy import  deepcopy

# ----------------------------------------------------------------------------------------------------------------------
# compute_performance
# ----------------------------------------------------------------------------------------------------------------------   
def compute_turbofan_performance(turbofan,state,center_of_gravity= [[0.0, 0.0,0.0]]):  
    ''' Computes the perfomrance of one turbofan
    
    Assumptions: 
    N/A

    Source:
    N/A
    
    Inputs:  
    turbofan             - turbofan data structure               [-]  
    state                - operating conditions data structure   [-]  
    fuel_line            - fuelline                              [-]
    center_of_gravity    - aircraft center of gravity            [m]

    Outputs:  
    total_thrust         - thrust of turbofan group              [N]
    total_momnet         - moment of turbofan group              [Nm]
    total_power          - power of turbofan group               [W] 
    stored_results_flag  - boolean for stored results            [-]     
    stored_propulsor_tag - name of turbojet with stored results  [-]
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                = state.conditions   
    noise_conditions          = conditions.noise[turbofan.tag] 
    turbofan_conditions       = conditions.energy[turbofan.tag] 
    rho                       = conditions.freestream.density
    U0                        = conditions.freestream.velocity
    ram                       = turbofan.ram
    inlet_nozzle              = turbofan.inlet_nozzle
    fan                       = turbofan.fan
    low_pressure_compressor   = turbofan.low_pressure_compressor
    high_pressure_compressor  = turbofan.high_pressure_compressor
    combustor                 = turbofan.combustor
    high_pressure_turbine     = turbofan.high_pressure_turbine
    low_pressure_turbine      = turbofan.low_pressure_turbine
    core_nozzle               = turbofan.core_nozzle
    fan_nozzle                = turbofan.fan_nozzle 
    bypass_ratio              = turbofan.bypass_ratio
    
    # unpack component conditions 
    ram_conditions          = turbofan_conditions[ram.tag]    
    inlet_nozzle_conditions = turbofan_conditions[inlet_nozzle.tag]
    fan_conditions          = turbofan_conditions[fan.tag]    
    lpc_conditions          = turbofan_conditions[low_pressure_compressor.tag]
    hpc_conditions          = turbofan_conditions[high_pressure_compressor.tag]
    combustor_conditions    = turbofan_conditions[combustor.tag]     
    lpt_conditions          = turbofan_conditions[low_pressure_turbine.tag]
    hpt_conditions          = turbofan_conditions[high_pressure_turbine.tag]
    core_nozzle_conditions  = turbofan_conditions[core_nozzle.tag]
    fan_nozzle_conditions   = turbofan_conditions[fan_nozzle.tag]    
 
    # Set the working fluid to determine the fluid properties
    ram.working_fluid = turbofan.working_fluid

    # Flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)

    # Link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature 
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)
    
    # Link the fan to the inlet nozzle
    fan_conditions.inputs.stagnation_temperature                      = inlet_nozzle_conditions.outputs.stagnation_temperature
    fan_conditions.inputs.stagnation_pressure                         = inlet_nozzle_conditions.outputs.stagnation_pressure
    fan_conditions.inputs.static_temperature                          = inlet_nozzle_conditions.outputs.static_temperature
    fan_conditions.inputs.static_pressure                             = inlet_nozzle_conditions.outputs.static_pressure
    fan_conditions.inputs.mach_number                                 = inlet_nozzle_conditions.outputs.mach_number  
    fan.working_fluid                                                 = turbofan.working_fluid
    
    # Flow through the fan
    compute_fan_performance(fan,fan_conditions,conditions)    

    # Link low pressure compressor to the inlet nozzle
    lpc_conditions.inputs.stagnation_temperature  = fan_conditions.outputs.stagnation_temperature
    lpc_conditions.inputs.stagnation_pressure     = fan_conditions.outputs.stagnation_pressure
    lpc_conditions.inputs.static_temperature      = fan_conditions.outputs.static_temperature
    lpc_conditions.inputs.static_pressure         = fan_conditions.outputs.static_pressure
    lpc_conditions.inputs.mach_number             = fan_conditions.outputs.mach_number  
    low_pressure_compressor.working_fluid         = turbofan.working_fluid
        
    # Flow through the low pressure compressor
    compute_compressor_performance(low_pressure_compressor,lpc_conditions,conditions)

    # Link the high pressure compressor to the low pressure compressor
    hpc_conditions.inputs.stagnation_temperature = lpc_conditions.outputs.stagnation_temperature
    hpc_conditions.inputs.stagnation_pressure    = lpc_conditions.outputs.stagnation_pressure
    hpc_conditions.inputs.static_temperature     = lpc_conditions.outputs.static_temperature
    hpc_conditions.inputs.static_pressure        = lpc_conditions.outputs.static_pressure
    hpc_conditions.inputs.mach_number            = lpc_conditions.outputs.mach_number  
    high_pressure_compressor.working_fluid       = low_pressure_compressor.working_fluid    
        
    # Flow through the high pressure compressor
    compute_compressor_performance(high_pressure_compressor,hpc_conditions,conditions)

    # Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature                = hpc_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = hpc_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = hpc_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = hpc_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = hpc_conditions.outputs.mach_number  
    combustor.working_fluid                                           = high_pressure_compressor.working_fluid     
        
    # Flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)

    # Link the shaft power output to the low pressure compressor
    try:
        shaft_power                                        = turbofan.Shaft_Power_Off_Take       
        shaft_power.inputs.mdhc                            = turbofan.compressor_nondimensional_massflow
        shaft_power.inputs.Tref                            = turbofan.reference_temperature
        shaft_power.inputs.Pref                            = turbofan.reference_pressure
        shaft_power.inputs.total_temperature_reference     = lpc_conditions.outputs.stagnation_temperature
        shaft_power.inputs.total_pressure_reference        = lpc_conditions.outputs.stagnation_pressure

        shaft_power(conditions)
        lpt_conditions.inputs.shaft_power_off_take   = shaft_power.outputs
    except:
        pass

    # Link the high pressure turbine to the combustor
    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number  
    hpt_conditions.inputs.compressor                = hpc_conditions.outputs 
    hpt_conditions.inputs.fan                       = fan_conditions.outputs
    hpt_conditions.inputs.bypass_ratio              = 0.0 #set to zero to ensure that fan not linked here 
    high_pressure_turbine.working_fluid             = combustor.working_fluid 
        
    # Flow through the high pressure turbine
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)

    # Link the low pressure turbine to the high pressure turbine
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure  
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number    
    lpt_conditions.inputs.compressor                 = lpc_conditions.outputs 
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.fan                        = fan_conditions.outputs  
    lpt_conditions.inputs.bypass_ratio               = bypass_ratio
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid  

    # Flow through the low pressure turbine
    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)

    # Link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = turbofan.working_fluid 
        
    # Flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions)

    # Link the dan nozzle to the fan
    fan_nozzle_conditions.inputs.stagnation_temperature     = fan_conditions.outputs.stagnation_temperature
    fan_nozzle_conditions.inputs.stagnation_pressure        = fan_conditions.outputs.stagnation_pressure
    fan_nozzle_conditions.inputs.static_temperature         = fan_conditions.outputs.static_temperature
    fan_nozzle_conditions.inputs.static_pressure            = fan_conditions.outputs.static_pressure  
    fan_nozzle_conditions.inputs.mach_number                = fan_conditions.outputs.mach_number   
    fan_nozzle.working_fluid                                = turbofan.working_fluid
        
    # Flow through the fan nozzle
    compute_expansion_nozzle_performance(fan_nozzle,fan_nozzle_conditions,conditions)
 
    # Link the thrust component to the fan nozzle
    turbofan_conditions.fan_nozzle_exit_velocity                        = fan_nozzle_conditions.outputs.velocity
    turbofan_conditions.fan_nozzle_area_ratio                           = fan_nozzle_conditions.outputs.area_ratio  
    turbofan_conditions.fan_nozzle_static_pressure                      = fan_nozzle_conditions.outputs.static_pressure
    turbofan_conditions.core_nozzle_area_ratio                          = core_nozzle_conditions.outputs.area_ratio 
    turbofan_conditions.core_nozzle_static_pressure                     = core_nozzle_conditions.outputs.static_pressure
    turbofan_conditions.core_nozzle_exit_velocity                       = core_nozzle_conditions.outputs.velocity  

    # Link the thrust component to the combustor
    turbofan_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio

    # Link the thrust component to the low pressure compressor 
    turbofan_conditions.total_temperature_reference              = lpc_conditions.outputs.stagnation_temperature
    turbofan_conditions.total_pressure_reference                 = lpc_conditions.outputs.stagnation_pressure 
    turbofan_conditions.bypass_ratio                             = bypass_ratio
    turbofan_conditions.flow_through_core                        = 1./(1.+bypass_ratio) #scaled constant to turn on core thrust computation
    turbofan_conditions.flow_through_fan                         = bypass_ratio/(1.+bypass_ratio) #scaled constant to turn on fan thrust computation        

    # Compute the thrust
    compute_thrust(turbofan,turbofan_conditions,conditions)

    # Compute forces and moments
    moment_vector              = 0*state.ones_row(3)
    thrust_vector              = 0*state.ones_row(3)
    thrust_vector[:,0]         =  turbofan_conditions.thrust[:,0]
    moment_vector[:,0]         =  turbofan.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1]         =  turbofan.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]         =  turbofan.origin[0][2]  -  center_of_gravity[0][2]
    M                          =  np.cross(moment_vector, thrust_vector)   
    moment                     = M 
    power                      = turbofan_conditions.power 
    turbofan_conditions.moment = moment

    # store data
    core_nozzle_res = Data(
                exit_static_temperature             = core_nozzle_conditions.outputs.static_temperature,
                exit_static_pressure                = core_nozzle_conditions.outputs.static_pressure,
                exit_stagnation_temperature         = core_nozzle_conditions.outputs.stagnation_temperature,
                exit_stagnation_pressure            = core_nozzle_conditions.outputs.static_pressure,
                exit_velocity                       = core_nozzle_conditions.outputs.velocity
            )

    fan_nozzle_res = Data(
                exit_static_temperature             = fan_nozzle_conditions.outputs.static_temperature,
                exit_static_pressure                = fan_nozzle_conditions.outputs.static_pressure,
                exit_stagnation_temperature         = fan_nozzle_conditions.outputs.stagnation_temperature,
                exit_stagnation_pressure            = fan_nozzle_conditions.outputs.static_pressure,
                exit_velocity                       = fan_nozzle_conditions.outputs.velocity
            )

    noise_conditions.turbofan.fan_nozzle    = fan_nozzle_res
    noise_conditions.turbofan.core_nozzle   = core_nozzle_res  
    noise_conditions.turbofan.fan           = None
    stored_results_flag                     = True
    stored_propulsor_tag                    = turbofan.tag 
    
    return thrust_vector,moment,power,stored_results_flag,stored_propulsor_tag 
    
def reuse_stored_turbofan_data(turbofan,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one turbofan for identical turbofans
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    turbofan             - turbofan data structure                [-]
    state                - operating conditions data structure   [-]  
    fuel_line            - fuelline                              [-]  
    total_thrust         - thrust of turbofan group              [N]
    total_power          - power of turbofan group               [W] 

    Outputs:  
    total_thrust         - thrust of turbofan group              [N]
    total_power          - power of turbofan group               [W] 
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                                      = state.conditions  
    conditions.energy[turbofan.tag]  = deepcopy(conditions.energy[stored_propulsor_tag])
    conditions.noise[turbofan.tag]   = deepcopy(conditions.noise[stored_propulsor_tag])
    
    # compute moment  
    moment_vector      = 0*state.ones_row(3)
    thrust_vector      = 0*state.ones_row(3)
    thrust_vector[:,0] = conditions.energy[turbofan.tag].thrust[:,0] 
    moment_vector[:,0] = turbofan.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1] = turbofan.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2] = turbofan.origin[0][2]  -  center_of_gravity[0][2]
    moment             = np.cross(moment_vector,thrust_vector)    
  
    power                                   = conditions.energy[turbofan.tag].power 
    conditions.energy[turbofan.tag].moment =  moment 
 
    return thrust_vector,moment,power    


# RCAIDE/Library/Missions/Common/Update/weights.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Update Weights
# ----------------------------------------------------------------------------------------------------------------------  
def weights(segment): 
    """ Updates the weight of the vehicle 
        
        Assumptions:
        N/A
        
        Inputs:
             segment.state.
                 numerics.time.integrate               [-]
                 conditions.weights.total_mass         [kg]
                 conditions.weights.vehicle_mass_rate  [kg/s]
                 conditions.freestream.gravity         [m/s^2]

                 
        Outputs: 
            segment.state.conditions.
                 weights.total_mass
                 frames.inertial.gravity_force_vector
      
        Properties Used:
        N/A
                    
    """ 
    
    # unpack
    conditions   = segment.state.conditions
    I            = segment.state.numerics.time.integrate  
    m0           = conditions.weights.total_mass[0,0]
    mdot_fuel    = conditions.weights.vehicle_mass_rate
    g            = conditions.freestream.gravity   
    
    networks = segment.analyses.energy.vehicle.networks
    for network in networks:
        if 'fuel_lines' in network:
            for fuel_line in network.fuel_lines:  
                fuel_line_results   = conditions.energy[fuel_line.tag] 
                for fuel_tank in fuel_line.fuel_tanks: 
                    fuel_line_results[fuel_tank.tag].mass[:,0]  =  fuel_line_results[fuel_tank.tag].mass[0,0]  + np.dot(I, -fuel_line_results[fuel_tank.tag].mass_flow_rate[:,0])   
            
    # calculate
    m = m0 + np.dot(I, -mdot_fuel )

    # weight
    W = m*g

    # pack
    conditions.weights.total_mass[1:,0]                  = m[1:,0]  
    conditions.frames.inertial.gravity_force_vector[:,2] = W[:,0]

    return
 


# RCAIDE/Methods/Performance/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .aircraft_aerodynamic_analysis     import aircraft_aerodynamic_analysis
from .estimate_take_off_field_length    import estimate_take_off_field_length
from .estimate_stall_speed              import estimate_stall_speed
from .compute_payload_range_diagram     import compute_payload_range_diagram
from .estimate_landing_field_length     import estimate_landing_field_length
from .find_take_off_weight_given_tofl   import find_take_off_weight_given_tofl
from .generate_V_n_diagram              import generate_V_n_diagram 
from .propeller_aerodynamic_analysis    import propeller_aerodynamic_analysis  



# RCAIDE/Compoments/Fuselages/Blended_Wing_Body_Fuselage.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports    
from .Fuselage import Fuselage

# python imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Blended_Wing_Body_Fuselage
# ----------------------------------------------------------------------------------------------------------------------  
class Blended_Wing_Body_Fuselage(Fuselage):
    """
    A blended wing body (BWB) fuselage design that smoothly integrates the wing and fuselage 
    into a single lifting body configuration.

    Attributes
    ----------
    tag : str
        Unique identifier for the BWB fuselage component, defaults to 'bwb_fuselage'
    
    aft_centerbody_area : float
        Cross-sectional area of the aft centerbody section in square meters
        
    aft_centerbody_taper : float
        Taper ratio of the aft centerbody section, defined as the ratio of tip 
        to root chord lengths
        
    cabin_area : float
        Total available cabin floor area in square meters

    Notes
    -----
    The blended wing body design offers several advantages over conventional tube-and-wing
    configurations:
    
    * Reduced wetted area leading to lower skin friction drag
    * Improved lift-to-drag ratio due to the lifting body design
    * Potential for increased internal volume and better weight distribution

    **Definitions**

    'Centerbody'
        The central section of the BWB that houses the passenger cabin and cargo hold
        
    'Aft Centerbody'
        The rear section of the centerbody that transitions into the outer wing sections

    See Also
    --------
    RCAIDE.Library.Components.Fuselages.Fuselage
        Base fuselage class that provides common functionality
    RCAIDE.Library.Components.Fuselages.Tube_Fuselage
        Conventional tube fuselage design for comparison
    """
    
    def __defaults__(self):
        """
        Sets the default values for the BWB fuselage component attributes.

        Notes
        -----
        This method initializes all required attributes with default values. Users should 
        modify these values based on their specific design requirements after instantiation.
        """      
          
        self.tag                   = 'bwb_fuselage'
        self.aft_centerbody_area   = 0.0
        self.aft_centerbody_taper  = 0.0
        self.cabin_area            = 0.0
        
    def compute_moment_of_inertia(self, center_of_gravity): 
        """
        Computes the moment of inertia tensor for the BWB fuselage.

        Parameters
        ----------
        center_of_gravity : array-like
            The (x, y, z) coordinates of the center of gravity about which to compute
            the moment of inertia

        Returns
        -------
        I : ndarray
            3x3 moment of inertia tensor in kg*m^2

        Notes
        -----
        Currently returns a zero matrix. This is a placeholder that should be implemented
        with actual BWB moment of inertia calculations.
        """
        I = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]]) 
        return I        


# RCAIDE/Library/Components/Energy/Sources/__init__.py
# 

"""
Energy sources module providing components for aircraft power generation and storage

This module contains implementations for various energy source components including
batteries, fuel tanks, and other power generation systems. These components serve
as the primary energy providers in aircraft propulsion systems.

See Also
--------
RCAIDE.Library.Components.Energy.Sources.Battery_Modules
    Battery system components and models
RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks
    Fuel storage and delivery systems
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from . import Battery_Modules
from . import Fuel_Tanks
from . import Solar_Panels


# RCAIDE/Methods/Noise/Metrics/EPNL_noise_metric.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  EPNL_noise_metric
# ----------------------------------------------------------------------------------------------------------------------        
def EPNL_noise_metric(PNLT):
    """This method calculates the effective perceived noise level (EPNL) based on a
    time history Perceived Noise Level with Tone Correction (PNLT).
     
    Assumptions:
        N/A

    Source:
        N/A

    Inputs:
        PNLT - Perceived Noise Level with Tone Correction  [PNLdB]
     
     Outputs:
        EPNL - Effective Perceived Noise Level             [EPNdB]
     
    Properties Used:
        N/A  
    """           
    # Maximum PNLT on the time history data    
    n_mic    = len(PNLT[0,:])
    PNLT_max = np.max(PNLT,axis=0)
    
    # Calculates the number of discrete points on the trajectory
    nsteps   = len(PNLT)    
    EPNL     = np.zeros(n_mic)
    for j in range(n_mic):
        # Finding the time duration for the noise history where PNL is higher than the maximum PNLT - 10 dB
        i = 0
        while PNLT[i][j]<=(PNLT_max[j]-10) and i<=nsteps:
            i = i+1
        t1 = i #t1 is the first time interval
        i  = i+1
    
        # Correction for PNLTM-10 when it falls outside the limit of the data
        if PNLT[nsteps-1][j]>=(PNLT_max[j]-10):
            t2=nsteps-2
        else:
            while i<=nsteps and PNLT[i][j]>=(PNLT_max[j]-10):
                i = i+1
            t2 = i-1 #t2 is the last time interval 
        
        # Calculates the integral of the PNLT which between t1 and t2 points
        sumation = 0
        for i in range (t1-1,t2+1):
            sumation = 10**(PNLT[i][j]/10)+sumation
            
        # Duration Correction calculation
        duration_correction = 10*np.log10(sumation)-PNLT_max[j]-13
                    
        # Final EPNL calculation
        EPNL[j] = PNLT_max[j]+duration_correction
    
    return EPNL   


# RCAIDE/Methods/Energy/Propulsors/Converters/Turbofan/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .append_turbofan_conditions                                 import append_turbofan_conditions 
from .compute_thurst                                             import compute_thrust
from .size_core                                                  import size_core 
from .compute_turbofan_performance                               import compute_turbofan_performance ,reuse_stored_turbofan_data
from .design_turbofan                                            import design_turbofan    



# RCAIDE/Compoments/Nacelles/Nacelle.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports    
from RCAIDE.Framework.Core              import Data 
from RCAIDE.Library.Components          import Component   
import scipy as sp
import numpy as np
 
# ---------------------------------------------------------------------------------------------------------------------- 
#  Nacelle
# ----------------------------------------------------------------------------------------------------------------------  
class Nacelle(Component):
    """
    Base class for aircraft engine nacelles providing core functionality for geometric 
    definition and analysis.

    Attributes
    ----------
    tag : str
        Unique identifier for the nacelle component, defaults to 'nacelle'
        
    origin : list
        3D coordinates [x, y, z] defining the nacelle's reference point, 
        defaults to [[0.0, 0.0, 0.0]]
        
    aerodynamic_center : list
        3D coordinates [x, y, z] of the nacelle's aerodynamic center, 
        defaults to [0.0, 0.0, 0.0]
        
    areas : Data
        Collection of nacelle area measurements
        
        - front_projected : float
            Frontal area, defaults to 0.0
        - side_projected : float
            Side profile area, defaults to 0.0
        - wetted : float
            Total wetted surface area, defaults to 0.0
            
    diameter : float
        Maximum diameter of the nacelle, defaults to 0.0
        
    inlet_diameter : float
        Diameter of the engine inlet, defaults to 0.0
        
    length : float
        Total length of the nacelle, defaults to 0.0
        
    orientation_euler_angles : list
        Rotation angles [roll, pitch, yaw] in radians, defaults to [0.0, 0.0, 0.0]
        
    flow_through : bool
        Flag indicating if nacelle has flow passing through it, defaults to True
        
    has_pylon : bool
        Flag indicating if nacelle is mounted on a pylon, defaults to True
        
    differential_pressure : float
        Pressure differential between internal and external flow, defaults to 0.0
        
    cowling_airfoil_angle : float
        Angle of the cowling lip airfoil section, defaults to 0.0

    Notes
    -----
    The nacelle class provides the foundation for engine installation design, including:
    
    * Geometric definition capabilities
    * Coordinate transformation utilities
    * Integration with propulsion system analysis
    
    **Major Assumptions**
    
    * Rigid body for structural analysis
    * Quasi-steady aerodynamics
    
    **Definitions**

    'Nacelle Frame'
        Local coordinate system with X out the nose, Z towards the ground, 
        and Y out the right side
        
    'Vehicle Frame'
        Aircraft coordinate system with X towards the tail, Z towards the ceiling, 
        and Y out the right wing

    See Also
    --------
    RCAIDE.Library.Components.Nacelles.Body_of_Revolution_Nacelle
        Implementation for axisymmetric nacelles
    RCAIDE.Library.Components.Nacelles.Stack_Nacelle
        Implementation for stacked segment nacelles
    """
    
    def __defaults__(self):
        """
        Sets default values for the nacelle attributes.
        """      
        self.tag                      = 'nacelle'
        self.origin                   = [[0.0,0.0,0.0]]
        self.aerodynamic_center       = [0.0,0.0,0.0]  
        self.areas                    = Data()
        self.areas.front_projected    = 0.0
        self.areas.side_projected     = 0.0
        self.areas.wetted             = 0.0 
        self.diameter                 = 0.0 
        self.inlet_diameter           = 0.0
        self.length                   = 0.0   
        self.orientation_euler_angles = [0.,0.,0.]    
        self.flow_through             = True
        self.has_pylon                = True
        self.differential_pressure    = 0.0    
        self.cowling_airfoil_angle    = 0.0

    def append_operating_conditions(self, segment, propulsor): 
        """
        Placeholder for adding operating conditions to the nacelle.

        Parameters
        ----------
        segment : Data
            Flight segment data
        propulsor : Data
            Propulsion system data
        """
        return

    def nac_vel_to_body(self):
        """
        Computes rotation matrix from nacelle velocity frame to body frame.

        Returns
        -------
        ndarray
            3x3 rotation matrix

        Notes
        -----
        Assumes two nacelle frames:
        * Vehicle frame describing location
        * Velocity frame for aerodynamic calculations
        """
        body2nacvel = self.body_to_nac_vel()
        r = sp.spatial.transform.Rotation.from_matrix(body2nacvel)
        r = r.inv()
        rot_mat = r.as_matrix()
        return rot_mat
    
    def body_to_nac_vel(self):
        """
        Computes rotation matrix from body frame to nacelle velocity frame.

        Returns
        -------
        ndarray
            3x3 rotation matrix

        Notes
        -----
        Transformation sequence:
        1. Body to vehicle frame
        2. Vehicle to nacelle vehicle frame
        3. Nacelle vehicle to nacelle velocity frame
        """
        # Go from body to vehicle frame
        body_2_vehicle = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()
        
        # Go from vehicle frame to nacelle vehicle frame
        rots = np.array(self.orientation_euler_angles) * 1. 
        vehicle_2_nac_vec = sp.spatial.transform.Rotation.from_rotvec(rots).as_matrix()        
        
        # Go from nacelle vehicle frame to nacelle velocity frame
        nac_vec_2_nac_vel = self.vec_to_vel()
        
        # Combine transformations
        rot1 = np.matmul(body_2_vehicle,vehicle_2_nac_vec)
        rot_mat = np.matmul(rot1,nac_vec_2_nac_vel) 
        return rot_mat    
    
    def vec_to_vel(self):
        """
        Computes rotation matrix from nacelle vehicle frame to nacelle velocity frame.

        Returns
        -------
        ndarray
            3x3 rotation matrix
        """
        rot_mat = sp.spatial.transform.Rotation.from_rotvec([0,np.pi,0]).as_matrix()
        return rot_mat
    
    
        


# RCAIDE/Library/Methods/Propulsors/Converters/Combustor/append_combustor_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_combustor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_combustor_conditions(combustor,segment,propulsor_conditions): 
    ones_row    = segment.state.ones_row 
    propulsor_conditions[combustor.tag]                           = Conditions() 
    propulsor_conditions[combustor.tag].inputs                    = Conditions() 
    propulsor_conditions[combustor.tag].inputs.nondim_mass_ratio  = ones_row(1)
    propulsor_conditions[combustor.tag].outputs                   = Conditions()
    return 



from .wing_planform import wing_planform
# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------
## @ingroup Methods-Geometry-Two_Dimensional-Cross_Section-Planform
def horizontal_tail_planform_raymer(horizontal_stabilizer, wing,  l_ht,c_ht):
    """Adjusts reference area before calling generic wing planform function to compute wing planform values.

    Assumptions:
    None

    Source:
    Raymer

    Inputs:
    horizontal_stabilizer                  [RCAIDE data structure]
    wing                                   [RCAIDE data structure]  (should be the main wing)
    l_ht                                   [m] length from wing mean aerodynamic chord (MAC) to horizontal stabilizer MAC
    c_ht                                   [-] horizontal tail coefficient (Raymer specific) .5 = Sailplane, .5 = homebuilt, 
                                               .7 = GA single engine, .8 = GA twin engine .5 = agricultural, .9 = twin turboprop, 
                                               .7 = flying boat, .7 = jet trainer, .4 = jet fighter, 1. = military cargo/bomber, 
                                               1. = jet transport

    Outputs:
    horizontal_stabilier.areas.reference   [m^2]
    Other changes to horizontal_stabilizer (see wing_planform)

    Properties Used:
    N/A
    """       
    
    horizontal_stabilizer.areas.reference = wing.chords.mean_aerodynamic*c_ht*wing.areas.reference/l_ht
    wing_planform(horizontal_stabilizer)
    
    return 0    



# RCAIDE/Framework/Analyses/Mission/Segments/Single_Point/Set_Speed_Set_Altitude_No_Propulsion.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Library.Methods                           import skip   
from RCAIDE.Framework.Core                            import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate       import Evaluate
from RCAIDE.Library.Mission                           import Common,Segments
 
# Package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Set_Speed_Set_Altitude_No_Propulsion
# ---------------------------------------------------------------------------------------------------------------------- 
class Set_Speed_Set_Altitude_No_Propulsion(Evaluate):
    """ This is a segment that is solved using a single point. A snapshot in time.
        We fix the speed and altitude. Throttle is solved from those.
    
        Assumptions:
        None
        
        Source:
        None
    """        
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # --------------------------------------------------------------------------------------------------------------
        #   User Inputs
        # --------------------------------------------------------------------------------------------------------------
        self.altitude                                = None
        self.air_speed                               = 10. * Units['km/hr']
        self.distance                                = 1.  * Units.km
        self.linear_acceleration_z                   = 0. # note that down is positive
        self.roll_rate                               = 0
        self.pitch_rate                              = 0
        self.yaw_rate                                = 0
        self.state.numerics.number_of_control_points = 1

        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------             
        initialize                         = self.process.initialize 
        initialize.expand_state            = skip
        initialize.differentials           = skip
        initialize.conditions              = Segments.Single_Point.Set_Speed_Set_Altitude_No_Propulsion.initialize_conditions 
        iterate                            = self.process.iterate
        iterate.conditions.differentials   = skip 
        iterate.conditions.weights         = Common.Update.weights
        iterate.conditions.planet_position = skip 
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.unknowns.orientation       = Common.Unpack_Unknowns.orientation
        
        return




# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Propulsion/integrated_propulsion.py
# 
# 
# Created:  Sep 2024, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Integrated Propulsion Weight 
# ----------------------------------------------------------------------------------------------------------------------
def integrated_propulsion(jet_engine_weight, engine_W_factor = 1.6):
    """ Calculate the weight of the entire propulsion system 
    
    Assumptions:
            The propulsion system is a fixed 60% greater than the dry engine alone. 
            The propulsion system includes the engines, engine exhaust, reverser, starting,
            controls, lubricating, and fuel systems. The nacelle and pylon weight are also
            part of this calculation.           
            
    Source: 
            N/A
            
    Inputs:
            compute_jet_engine_weight - dry weight of the engine                                             [kilograms]
            num_eng - total number of engines on the aircraft                                 [dimensionless]
            engine_W_factor - weight increase factor for entire integrated propulsion system [dimensionless]
    
    Outputs:
            weight - weight of the full propulsion system                                     [kilograms]
        
    Properties Used:
            N/A
    """   
    
    weight = jet_engine_weight * engine_W_factor
    
    return weight
    


# RCAIDE/Library/Components/Propulsors/ICE_Propeller.py
# 
#  
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from .                import Propulsor  
from RCAIDE.Library.Methods.Propulsors.ICE_Propulsor.unpack_ice_propeller_unknowns   import unpack_ice_propeller_unknowns
from RCAIDE.Library.Methods.Propulsors.ICE_Propulsor.pack_ice_propeller_residuals    import pack_ice_propeller_residuals
from RCAIDE.Library.Methods.Propulsors.ICE_Propulsor.append_ice_propeller_conditions import append_ice_propeller_conditions
from RCAIDE.Library.Methods.Propulsors.ICE_Propulsor.compute_ice_performance         import compute_ice_performance, reuse_stored_ice_data
from RCAIDE.Library.Methods.Propulsors.ICE_Propulsor.append_ice_residual_and_unknown import  append_ice_residual_and_unknown
 

# ---------------------------------------------------------------------------------------------------------------------- 
# ICE_Propeller
# ---------------------------------------------------------------------------------------------------------------------- 
class ICE_Propeller(Propulsor):
    """
    A propulsion system class that combines an internal combustion engine with a fixed-pitch propeller.
    
    Attributes
    ----------
    tag : str
        Identifier for the propulsion system, defaults to 'ice_propeller'
    
    active_fuel_tanks : list
        List with names of active fuel tanks. Default is None.
        
    engine : None or Engine
        The internal combustion engine component
        
    propeller : None or Propeller
        The fixed-pitch propeller component
    
    Notes
    -----
    This class models a conventional propulsion system that pairs an internal 
    combustion engine with a fixed-pitch propeller. Unlike constant-speed propellers, 
    the blade pitch remains fixed during operation, making RPM directly dependent 
    on throttle setting and flight conditions.
    
    Key characteristics:
    - Simpler mechanical design than constant-speed systems
    - RPM varies with airspeed and power setting
    - Optimized for a specific flight condition
    - Lower cost and maintenance requirements
    
    **Definitions**

    'Fixed-Pitch Propeller'
        A propeller with blades set at a fixed angle, optimized for 
        a specific flight regime (typically cruise)
    
    'Blade Angle'
        The angle between the blade's chord line and the plane of rotation,
        measured at a specific radial station
    
    See Also
    --------
    RCAIDE.Library.Components.Propulsors.Propulsor
    RCAIDE.Library.Components.Propulsors.Constant_Speed_ICE_Propeller
    """
    def __defaults__(self):    
        # setting the default values
        self.tag                          = 'ice_propeller'   
        self.active_fuel_tanks            = None
        self.engine                       = None
        self.propeller                    = None
        self.engine_diameter              = 0.0      
        self.engine_length                = 0.0
        self.engine_mass                  = 0.0

    def append_operating_conditions(self,segment):
        """
        Appends operating conditions to the segment.
        """
        append_ice_propeller_conditions(self,segment)
        return

    def unpack_propulsor_unknowns(self,segment):  
        """
        Unpacks propulsor unknowns from the segment.
        """
        unpack_ice_propeller_unknowns(self,segment)
        return 

    def pack_propulsor_residuals(self,segment): 
        """
        Packs propulsor residuals into the segment.
        """
        pack_ice_propeller_residuals(self,segment)
        return

    def append_propulsor_unknowns_and_residuals(self,segment):
        """
        Appends propulsor unknowns and residuals to the segment.
        """
        append_ice_residual_and_unknown(self,segment)
        return    
    
    def compute_performance(self,state,center_of_gravity = [[0, 0, 0]]):
        """
        Computes propulsor performance including thrust, moment, and power.
        """
        thrust,moment,power,stored_results_flag,stored_propulsor_tag =  compute_ice_performance(self,state,center_of_gravity)
        return thrust,moment,power,stored_results_flag,stored_propulsor_tag
    
    def reuse_stored_data(ICE_prop,state,network,stored_propulsor_tag,center_of_gravity = [[0, 0, 0]]):
        """
        Reuses stored propulsor data for performance calculations.
        """
        thrust,moment,power  = reuse_stored_ice_data(ICE_prop,state,network,stored_propulsor_tag,center_of_gravity)
        return thrust,moment,power 


# RCAIDE/Library/Missions/Common/Unpack_Unknowns/control_surfaces.py
# 
# 
# Created:  Jul 2023, M. Clarke
import RCAIDE

# ----------------------------------------------------------------------------------------------------------------------
#  Unpack Unknowns
# ----------------------------------------------------------------------------------------------------------------------
def control_surfaces(segment):
    assigned_control_variables   = segment.assigned_control_variables
    control_surfaces             = segment.state.conditions.control_surfaces
    
    for analysis in segment.analyses:
        if analysis !=  None: 
            if 'vehicle' in analysis: 
                wings =  analysis.vehicle.wings 
                # loop through wings on aircraft
                for wing in wings:
                    # Elevator Control
                    if assigned_control_variables.elevator_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Elevator:
                                num_elev_ctrls = len(assigned_control_variables.elevator_deflection.assigned_surfaces)
                                for i in range(num_elev_ctrls):   
                                    for j in range(len(assigned_control_variables.elevator_deflection.assigned_surfaces[i])):
                                        elevator_name = assigned_control_variables.elevator_deflection.assigned_surfaces[i][j]
            
                                        # set deflection on vehicle
                                        wing.control_surfaces[elevator_name].deflection = segment.state.unknowns["elevator_" + str(i)]
            
                                        # set deflection in results data structure
                                        control_surfaces.elevator.deflection  = segment.state.unknowns["elevator_" + str(i)]
                            
                    # Slat Control
                    if assigned_control_variables.slat_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Slat:
                                num_slat_ctrls = len(assigned_control_variables.slat_deflection.assigned_surfaces)
                                for i in range(num_slat_ctrls):   
                                    for j in range(len(assigned_control_variables.slat_deflection.assigned_surfaces[i])):
                                        slat_name = assigned_control_variables.slat_deflection.assigned_surfaces[i][j]
            
                                        # set deflection on vehicle
                                        wing.control_surfaces[slat_name].deflection = segment.state.unknowns["slat_" + str(i)]
            
                                        # set deflection in results data structure
                                        control_surfaces.slat.deflection  = segment.state.unknowns["slat_" + str(i)]
            
            
                    # Rudder Control
                    if assigned_control_variables.rudder_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Rudder:
                                num_rud_ctrls = len(assigned_control_variables.rudder_deflection.assigned_surfaces)
                                for i in range(num_rud_ctrls):
                                    for j in range(len(assigned_control_variables.rudder_deflection.assigned_surfaces[i])):
                                        rudder_name = assigned_control_variables.rudder_deflection.assigned_surfaces[i][j]
            
                                        # set deflection on vehicle
                                        wing.control_surfaces[rudder_name].deflection = segment.state.unknowns["rudder_" + str(i)]
            
                                        # set deflection in results data structure
                                        control_surfaces.rudder.deflection  = segment.state.unknowns["rudder_" + str(i)]
            
                    # flap Control
                    if assigned_control_variables.flap_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Flap:
                                num_flap_ctrls = len(assigned_control_variables.flap_deflection.assigned_surfaces)
                                for i in range(num_flap_ctrls):   
                                    for j in range(len(assigned_control_variables.flap_deflection.assigned_surfaces[i])):
                                        flap_name = assigned_control_variables.flap_deflection.assigned_surfaces[i][j]
            
                                        # set deflection on vehicle
                                        wing.control_surfaces[flap_name].deflection = segment.state.unknowns["flap_" + str(i)]
            
                                        # set deflection in results data structure
                                        control_surfaces.flap.deflection  = segment.state.unknowns["flap_" + str(i)]
            
                    # Aileron Control
                    if assigned_control_variables.aileron_deflection.active:
                        for control_surface in wing.control_surfaces:
                            if type(control_surface) == RCAIDE.Library.Components.Wings.Control_Surfaces.Aileron:
                                num_aile_ctrls = len(assigned_control_variables.aileron_deflection.assigned_surfaces)
                                for i in range(num_aile_ctrls):   
                                    for j in range(len(assigned_control_variables.aileron_deflection.assigned_surfaces[i])):
                                        aileron_name = assigned_control_variables.aileron_deflection.assigned_surfaces[i][j]
            
                                        # set deflection on vehicle
                                        wing.control_surfaces[aileron_name].deflection = segment.state.unknowns["aileron_" + str(i)]
            
                                        # set deflection in results data structure
                                        control_surfaces.aileron.deflection  = segment.state.unknowns["aileron_" + str(i)]


    
    return


# RCAIDE/Library/Attributes/Gases/CO2.py
# 
#
# Created:  Mar 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------  
#  Imports
# ----------------------------------------------------------------------------------------------------------------------  
 
from .Gas import Gas  

# ----------------------------------------------------------------------------------------------------------------------  
# Steam Class
# ----------------------------------------------------------------------------------------------------------------------   
class Steam(Gas):
    """Holds constants and functions that compute gas properties for steam. 
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """ 
        self.tag                   ='steam'
        self.molecular_mass        = 18.           # kg/kmol
        self.gas_specific_constant = 461.889       # m^2/s^2-K, specific gas constant
        self.composition.H2O       = 1.0

    def compute_density(self,T=300,p=101325):
        """Computes steam density given temperature and pressure 
        
        Assumptions:
            Ideal gas
            
        Source:
            None
    
        Args:
            self       : steam                 [unitless]
            T (float)  : temperature           [K]
            P (float)  : pressure              [Pa]
            
        Returns:
            rho (float): density               [kg/m^3]       
        """        
        return p/(self.gas_specific_constant*T)

    def compute_speed_of_sound(self,T=300,p=101325,variable_gamma=False):
        """Computes speed of sound given temperature and pressure

        Assumptions:
            Ideal gas with gamma = 1.33 if variable gamma is False

        Source:
            None 

        Args:
            self                     : steam         [unitless]
            T (float)                : temperature   [K]    
            p (float)                : Pressure      [Pa]      
            variable_gamma (boolean) :               [unitless]

        Returns:
            a (float)                : speed of sound [m/s] 
        """        
        if variable_gamma:
            g = self.compute_gamma(T,p)
        else:
            g = 1.33

        return (g*self.gas_specific_constant*T)**0.5 

    def compute_cp(self,T=300,p=101325):
        """Computes Cp by 3rd-order polynomial data fit: cp(T) = c1*T^3 + c2*T^2 + c3*T + c4
             
        Assumptions:
            300 K < T < 1500 K
            
        Source:
            Unknown, possibly Combustion Technologies for a Clean Environment 
            (Energy, Combustion and the Environment), Jun 15, 1995, Carvalhoc
    
        Args:
            self      : steam             [unitless]
            T (float) : temperature       [K]
            P (float) : pressure          [Pa]
            
        Returns:
            cp (float): specfic heat capacity [J/(kg K)]         
        """   
        c = [5E-9, -.0001,  .9202, 1524.7]
        cp = c[0]*T**3 + c[1]*T**2 + c[2]*T + c[3]

        return cp

    def compute_gamma(self,T=300,p=101325):
        """Gives constant gamma of 1.33 
        
        Assumptions:
             233 K < T < 1273 K 
            
        Source:
            None
    
        Args:
            self      : steam         [unitless]
            T (float) : temperature   [K]
            P (float) : pressure      [Pa]
            
        Returns: 
            g  (float): gamma         [unitless] 
        """           
        g = 1.33   
        return g

    def compute_absolute_viscosity(self,T=300,p=101325):
        """Returns a constant absolute viscosity of 1e-6 
        
        Assumptions:
            None
            
        Source:
            None
    
        Args:
            self      : steam                 [unitless]
            T (float) : temperature           [K]
            P (float) : pressure              [Pa]
            
        Returns:
            mu (float): absolute viscosity    [kg/(m-s)]       
        """ 
        mu =1E-6 
        return mu 


## @ingroup Methods-Aerodynamics-Fidelity_Zero-Lift
# compute_flap_lift.py
#
# Created:  Dec 2013, A. Varyar
# Modified: Feb 2014, T. Orra
#           Jan 2016, E. Botero         

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
from   RCAIDE import  * 
from   RCAIDE.Framework.Core import Units
import numpy  as np

# ----------------------------------------------------------------------
#  compute_flap_lift
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Fidelity_Zero-Lift
def compute_flap_lift(t_c,flap_type,flap_chord,flap_angle,sweep,wing_Sref,wing_affected_area):
    """Computes the increase of lift due to trailing edge flap deployment

    Assumptions:
    None

    Source:
    Unknown

    Inputs:
    t_c                 (wing thickness ratio)                 [Unitless]
    flap_type                                                  [string]
    flap_c_chord        (flap chord as fraction of wing chord) [Unitless]
    flap_angle          (flap deflection)                      [radians]
    sweep               (Wing sweep angle)                     [radians]
    wing_Sref           (Wing reference area)                  [m^2]
    wing_affected_area  (Wing area affected by flaps)          [m^2]

    Outputs:
    dcl_max_flaps       (Lift coefficient increase)            [Unitless]

    Properties Used:
    N/A
    """          

    #unpack
    tc_r  = t_c
    fc    = flap_chord * 100.
    fa    = flap_angle / Units.deg
    Swf   = wing_affected_area
    sweep = sweep

    # Basic increase in CL due to flap
    dmax_ref= -4E-05*tc_r**4 + 0.0014*tc_r**3 - 0.0093*tc_r**2 + 0.0436*tc_r + 0.9734

    # Corrections for flap type
    if flap_type == None:
        dmax_ref = 0.
    elif flap_type.upper() == 'single_slotted'.upper():
        dmax_ref = dmax_ref * 0.93
    elif flap_type.upper() == 'triple_slotted'.upper():
        dmax_ref = dmax_ref * 1.08

    # Chord correction
    Kc =  0.0395*fc    + 0.0057

    # Deflection correction
    Kd = -1.7857E-04*fa**2 + 2.9214E-02*fa - 1.4000E-02

    # Sweep correction
    Ksw = (1 - 0.08 * (np.cos(sweep))**2) * (np.cos(sweep)) ** 0.75

    # Applying corrections
    dmax_flaps = Kc * Kd * Ksw * dmax_ref

    # Final CL increment due to flap
    dcl_max_flaps = dmax_flaps  *  Swf / wing_Sref

    return dcl_max_flaps


# RCAIDE/Compoments/Landing_Gear/Landing_Gear.py
# 
# Created:  Nov 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports     
from RCAIDE.Library.Components import Component    

# ---------------------------------------------------------------------------------------------------------------------- 
#  Landing_Gear
# ----------------------------------------------------------------------------------------------------------------------  
class Landing_Gear(Component):
    """
    Base class for aircraft landing gear components providing core functionality for 
    both main and nose gear configurations.

    Attributes
    ----------
    tag : str
        Unique identifier for the landing gear component, defaults to 'landing_gear'
        
    tire_diameter : float
        Diameter of the landing gear tire, defaults to 0
        
    strut_length : float
        Length of the landing gear strut assembly, defaults to 0
        
    units : int
        Number of landing gear units in the assembly, defaults to 0
        
    gear_extended : bool
        Flag indicating whether the landing gear is in extended position, 
        defaults to False
        
    wheels : int
        Number of wheels per landing gear unit, defaults to 0

    Notes
    -----
    This class serves as the foundation for specific landing gear types, providing:
    
    * Basic geometric properties
    * Configuration parameters
    * State tracking (extended/retracted)
    
    **Major Assumptions**
    
    * Landing gear components are rigid bodies
    * Tire and strut properties are uniform within each unit
    * Gear state is binary (either fully extended or fully retracted)
    
    **Definitions**

    'Strut'
        The main structural member of the landing gear that absorbs landing loads
        and supports the wheel assembly
        
    'Unit'
        A complete landing gear assembly including strut, wheels, and associated 
        mechanisms

    See Also
    --------
    RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear
        Implementation for main landing gear
    RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear
        Implementation for nose landing gear
    """

    def __defaults__(self):
        """
        Sets default values for the landing gear attributes.
        """
        self.tag            = 'landing_gear'   
        self.tire_diameter  = 0 
        self.strut_length   = 0 
        self.units          = 0 
        self.gear_extended  = False
        self.wheels         = 0     


# RCAIDE/Framework/External_Interfaces/OpenVSP/vsp_wing.py

# Created:  Jun 2018, T. St Francis
# Modified: Aug 2018, T. St Francis
#           Jan 2020, T. MacDonald
#           Jul 2020, E. Botero
#           May 2021, E. Botero
#           Feb 2022, M. Cunningham

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units , Data 
from RCAIDE.Library.Components.Airfoils.Airfoil import Airfoil
from RCAIDE.Library.Methods.Geometry.Planform import wing_planform, wing_segmented_planform

import numpy as np
import string
try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
# This enforces lowercase names
chars = string.punctuation + string.whitespace
t_table = str.maketrans( chars          + string.ascii_uppercase ,
                         '_'*len(chars) + string.ascii_lowercase )

# ----------------------------------------------------------------------------------------------------------------------  
#  vsp read wing
# ----------------------------------------------------------------------------------------------------------------------  
def read_vsp_wing(wing_id, units_type='SI', write_airfoil_file=True, use_scaling=True):
    """This reads an OpenVSP wing vehicle geometry and writes it into a RCAIDE wing format.

    Assumptions:
    1. OpenVSP wing is divided into segments ("XSecs" in VSP).
    2. Written for OpenVSP 3.21.1

    Source:
    N/A

    Inputs:
    1. VSP 10-digit geom ID for wing.
    2. units_type set to 'SI' (default) or 'Imperial'.
    3. Boolean for whether or not to write an airfoil file(default = True).
    4. Boolean for whether or not to use the scaling from OpenVSP (default = True).

    Outputs:
    Writes RCAIDE wing object, with these geometries, from VSP:
    	Wings.Wing.    (* is all keys)
    		origin                                  [m] in all three dimensions
    		spans.projected                         [m]
    		chords.root                             [m]
    		chords.tip                              [m]
    		aspect_ratio                            [-]
    		sweeps.quarter_chord                    [radians]
    		twists.root                             [radians]
    		twists.tip                              [radians]
    		thickness_to_chord                      [-]
    		dihedral                                [radians]
    		symmetric                               <boolean>
    		tag                                     <string>
    		areas.reference                         [m^2]
    		areas.wetted                            [m^2]
    		Segments.
    		  tag                                   <string>
    		  twist                                 [radians]
    		  percent_span_location                 [-]  .1 is 10%
    		  root_chord_percent                    [-]  .1 is 10%
    		  dihedral_outboard                     [radians]
    		  sweeps.quarter_chord                  [radians]
    		  thickness_to_chord                    [-]
    		  airfoil                               <NACA 4-series, 6 series, or airfoil file>

    Properties Used:
    N/A
    """

    # Check if this is vertical tail, this seems like a weird first step but it's necessary
    # Get the initial rotation to get the dihedral angles
    x_rot = vsp.GetParmVal( wing_id,'X_Rotation','XForm')
    y_rot = vsp.GetParmVal( wing_id,'Y_Rotation','XForm')
    if  abs(x_rot) >=70:
        wing = RCAIDE.Library.Components.Wings.Vertical_Tail()
        wing.vertical = True
        sign = (np.sign(x_rot))
        x_rot = (sign*90 - sign*x_rot) * Units.deg
    else:
        # Instantiate a wing
        wing = RCAIDE.Library.Components.Wings.Wing()
        x_rot =  x_rot  * Units.deg

    y_rot =  y_rot  * Units.deg

    # Set the units
    if units_type == 'SI':
        units_factor = Units.meter * 1.
    elif units_type == 'imperial':
        units_factor = Units.foot * 1.
    elif units_type == 'inches':
        units_factor = Units.inch * 1.

    # Apply a tag to the wing
    if vsp.GetGeomName(wing_id):
        tag = vsp.GetGeomName(wing_id)
        tag = tag.translate(t_table)
        wing.tag = tag
    else:
        wing.tag = 'winggeom'

    if use_scaling:
        scaling       = vsp.GetParmVal(wing_id, 'Scale', 'XForm')
    else:
        scaling       = 1.
    units_factor      = units_factor*scaling

    # Top level wing parameters
    # Wing origin
    wing.origin[0][0] = vsp.GetParmVal(wing_id, 'X_Location', 'XForm') * units_factor
    wing.origin[0][1] = vsp.GetParmVal(wing_id, 'Y_Location', 'XForm') * units_factor
    wing.origin[0][2] = vsp.GetParmVal(wing_id, 'Z_Location', 'XForm') * units_factor

    # Wing Symmetry
    sym_planar = vsp.GetParmVal(wing_id, 'Sym_Planar_Flag', 'Sym')
    sym_origin = vsp.GetParmVal(wing_id, 'Sym_Ancestor_Origin_Flag', 'Sym')

    # Check for symmetry
    if sym_planar == 2. and sym_origin == 1.: #origin at wing, not vehicle
        wing.symmetric = True
    else:
        wing.symmetric = False

    #More top level parameters
    total_proj_span      = vsp.GetParmVal(wing_id, 'TotalProjectedSpan', 'WingGeom') * units_factor
    wing.aspect_ratio    = vsp.GetParmVal(wing_id, 'TotalAR', 'WingGeom')
    wing.areas.reference = vsp.GetParmVal(wing_id, 'TotalArea', 'WingGeom') * units_factor**2
    wing.spans.projected = total_proj_span

    # Check if this is a single segment wing
    xsec_surf_id      = vsp.GetXSecSurf(wing_id, 0)   # This is how VSP stores surfaces.
    x_sec_1           = vsp.GetXSec(xsec_surf_id, 1)

    if vsp.GetNumXSec(xsec_surf_id) == 2:
        single_seg = True
    else:
        single_seg = False

    segment_num = vsp.GetNumXSec(xsec_surf_id) # Get number of segments

    span_sum         = 0.				# Non-projected.
    proj_span_sum    = 0.				# Projected.
    segment_spans    = [None] * (segment_num) 	        # Non-projected.
    segment_dihedral = [None] * (segment_num)
    segment_sweeps_quarter_chord = [None] * (segment_num)

    # Necessary wing segment definitions start at XSec_1 (XSec_0 exists mainly to hold the root airfoil)
    xsec_surf_id = vsp.GetXSecSurf(wing_id, 0)
    x_sec = vsp.GetXSec(xsec_surf_id, 1)
    chord_parm = vsp.GetXSecParm(x_sec,'Root_Chord')
    root_chord = vsp.GetParmVal(chord_parm) * units_factor

    # -------------
    # Wing segments
    # -------------

    if single_seg == False:

        # Convert VSP XSecs to RCAIDE segments. (Wing segments are defined by outboard sections in VSP, but inboard sections in RCAIDE.)
        for i in range(1, segment_num+1):
            # XSec airfoil
            jj = i-1  # Airfoil index i-1 because VSP airfoils and sections are one index off relative to RCAIDE.

            segment = RCAIDE.Library.Components.Wings.Segments.Segment()
            segment.tag                   = 'Section_' + str(i)
            thick_cord                    = vsp.GetParmVal(wing_id, 'ThickChord', 'XSecCurve_' + str(jj))
            segment.thickness_to_chord    = thick_cord	# Thick_cord stored for use in airfoil, below.
            if i!=segment_num:
                segment_root_chord    = vsp.GetParmVal(wing_id, 'Root_Chord', 'XSec_' + str(i)) * units_factor
            else:
                segment_root_chord    = 0.0
            segment.root_chord_percent    = segment_root_chord / root_chord
            segment.percent_span_location = proj_span_sum / (total_proj_span/(1+wing.symmetric))
            segment.twist                 = vsp.GetParmVal(wing_id, 'Twist', 'XSec_' + str(jj)) * Units.deg +  y_rot

            if i==1:
                wing.thickness_to_chord = thick_cord

            if i < segment_num:      # This excludes the tip xsec, but we need a segment in RCAIDE to store airfoil.
                sweep     = vsp.GetParmVal(wing_id, 'Sweep', 'XSec_' + str(i)) * Units.deg
                sweep_loc = vsp.GetParmVal(wing_id, 'Sweep_Location', 'XSec_' + str(i))
                AR        = 2*vsp.GetParmVal(wing_id, 'Aspect', 'XSec_' + str(i))
                taper     = vsp.GetParmVal(wing_id, 'Taper', 'XSec_' + str(i))

                segment_sweeps_quarter_chord[i] = convert_sweep(sweep,sweep_loc,0.25,AR,taper)
                segment.sweeps.quarter_chord    = segment_sweeps_quarter_chord[i]  # Used again, below

                # Used for dihedral computation, below.
                segment_dihedral[i]	      = vsp.GetParmVal(wing_id, 'Dihedral', 'XSec_' + str(i)) * Units.deg  + x_rot
                segment.dihedral_outboard     = segment_dihedral[i]

                segment_spans[i] 	      = vsp.GetParmVal(wing_id, 'Span', 'XSec_' + str(i)) * units_factor
                proj_span_sum += segment_spans[i] * np.cos(segment_dihedral[i])
                span_sum      += segment_spans[i]
            else:
                segment.root_chord_percent    = (vsp.GetParmVal(wing_id, 'Tip_Chord', 'XSec_' + str(i-1))) * units_factor /root_chord


            xsec_id          = str(vsp.GetXSec(xsec_surf_id, jj))
            airfoil          = Airfoil()  # TO DO : ADD MORE AIRFOILS 
            airfoil.geometry = Data()
            if vsp.GetXSecShape(xsec_id) == vsp.XS_FOUR_SERIES: 	# XSec shape: NACA 4-series
                camber = vsp.GetParmVal(wing_id, 'Camber', 'XSecCurve_' + str(jj))

                if camber == 0.:
                    camber_loc = 0.
                else:
                    camber_loc = vsp.GetParmVal(wing_id, 'CamberLoc', 'XSecCurve_' + str(jj))

                airfoil.geometry.thickness_to_chord = thick_cord
                camber_round               = int(np.around(camber*100))
                camber_loc_round           = int(np.around(camber_loc*10))
                thick_cord_round           = int(np.around(thick_cord*100))
                airfoil.tag                = 'NACA ' + str(camber_round) + str(camber_loc_round) + str(thick_cord_round)

            elif vsp.GetXSecShape(xsec_id) == vsp.XS_SIX_SERIES: 	# XSec shape: NACA 6-series
                thick_cord_round = int(np.around(thick_cord*100))
                a_value          = vsp.GetParmVal(wing_id, 'A', 'XSecCurve_' + str(jj))
                ideal_CL         = int(np.around(vsp.GetParmVal(wing_id, 'IdealCl', 'XSecCurve_' + str(jj))*10))
                series_vsp       = int(vsp.GetParmVal(wing_id, 'Series', 'XSecCurve_' + str(jj)))
                series_dict      = {0:'63',1:'64',2:'65',3:'66',4:'67',5:'63A',6:'64A',7:'65A'} # VSP series values.
                series           = series_dict[series_vsp]
                airfoil.tag      = 'NACA ' + series + str(ideal_CL) + str(thick_cord_round) + ' a=' + str(np.around(a_value,1))


            elif vsp.GetXSecShape(xsec_id) == vsp.XS_FILE_AIRFOIL:	# XSec shape: 12 is type AF_FILE
                airfoil.geometry.thickness_to_chord = thick_cord
                # VSP airfoil API calls get coordinates and write files with the final argument being the fraction of segment position, regardless of relative spans.
                # (Write the root airfoil with final arg = 0. Write 4th airfoil of 5 segments with final arg = .8)

            if write_airfoil_file==True:
                vsp.WriteSeligAirfoil(str(wing.tag) + '_airfoil_XSec_' + str(jj) +'.dat', wing_id, float(jj/segment_num))
                airfoil.coordinate_file    = str(wing.tag) + '_airfoil_XSec_' + str(jj) +'.dat'
                airfoil.tag                = 'airfoil'

                segment.append_airfoil(airfoil)

            wing.segments.append(segment)

        # Wing dihedral
        proj_span_sum_alt = 0.
        span_sum_alt      = 0.
        sweeps_sum        = 0.

        for ii in range(1, segment_num):
            span_sum_alt      += segment_spans[ii]
            proj_span_sum_alt += segment_spans[ii] * np.cos(segment_dihedral[ii])  # Use projected span to find total wing dihedral.
            sweeps_sum        += segment_spans[ii] * np.tan(segment_sweeps_quarter_chord[ii])

        wing.dihedral              = np.arccos(proj_span_sum_alt / span_sum_alt)
        wing.sweeps.quarter_chord  = -np.arctan(sweeps_sum / span_sum_alt)  # Minus sign makes it positive sweep.

        # Add a tip segment, all values are zero except the tip chord
        tc = vsp.GetParmVal(wing_id, 'Tip_Chord', 'XSec_' + str(segment_num-1)) * units_factor

        # Chords
        wing.chords.root              = vsp.GetParmVal(wing_id, 'Tip_Chord', 'XSec_0') * units_factor
        wing.chords.tip               = tc
        wing.chords.mean_geometric    = wing.areas.reference / wing.spans.projected

        # Just double calculate and fix things:
        wing = wing_segmented_planform(wing)


    else:
        # Single segment

        # Get ID's
        x_sec_1_dih_parm       = vsp.GetXSecParm(x_sec_1,'Dihedral')
        x_sec_1_sweep_parm     = vsp.GetXSecParm(x_sec_1,'Sweep')
        x_sec_1_sweep_loc_parm = vsp.GetXSecParm(x_sec_1,'Sweep_Location')
        x_sec_1_taper_parm     = vsp.GetXSecParm(x_sec_1,'Taper')
        x_sec_1_rc_parm        = vsp.GetXSecParm(x_sec_1,'Root_Chord')
        x_sec_1_tc_parm        = vsp.GetXSecParm(x_sec_1,'Tip_Chord')
        x_sec_1_t_parm         = vsp.GetXSecParm(x_sec_1,'ThickChord')

        # Calcs
        sweep     = vsp.GetParmVal(x_sec_1_sweep_parm) * Units.deg
        sweep_loc = vsp.GetParmVal(x_sec_1_sweep_loc_parm)
        taper     = vsp.GetParmVal(x_sec_1_taper_parm)
        c_4_sweep = convert_sweep(sweep,sweep_loc,0.25,wing.aspect_ratio,taper)

        # Pull and pack
        wing.sweeps.quarter_chord  = c_4_sweep
        wing.taper                 = taper
        wing.dihedral              = vsp.GetParmVal(x_sec_1_dih_parm) * Units.deg + x_rot
        wing.chords.root           = vsp.GetParmVal(x_sec_1_rc_parm)* units_factor
        wing.chords.tip            = vsp.GetParmVal(x_sec_1_tc_parm) * units_factor
        wing.chords.mean_geometric = wing.areas.reference / wing.spans.projected
        wing.thickness_to_chord    = vsp.GetParmVal(x_sec_1_t_parm)

        # Just double calculate and fix things:
        wing = wing_planform(wing)
 

    # Twists
    wing.twists.root      = vsp.GetParmVal(wing_id, 'Twist', 'XSec_0') * Units.deg +  y_rot
    wing.twists.tip       = vsp.GetParmVal(wing_id, 'Twist', 'XSec_' + str(segment_num-1)) * Units.deg +  y_rot

    # check if control surface (sub surfaces) are defined
    tags                 = []
    LE_flags             = []
    U_starts             = []
    U_ends               = []
    chord_fractions      = []
    
    # determine the number of segments and where the breaks are
    if len(wing.segments.keys())>0:
        N = len(wing.segments.keys())-1
        
        # Do a for loop to find the location of breaks
        breaks = []
        for seg in wing.segments:
            breaks.append(seg.percent_span_location)
        
    else:
        N = 1
        # Location of breaks
        breaks = [0.,1.]     

    num_cs = vsp.GetNumSubSurf(wing_id)

    # loop through wing and get all control surface parameters
    for cs_idx in range(num_cs):
        cs_id   = vsp.GetSubSurf(wing_id,cs_idx)
        param_names = vsp.GetSubSurfParmIDs(cs_id)
        tags.append(vsp.GetSubSurfName(cs_id))
        for p_idx in range(len(param_names)):
            if 'LE_Flag' == vsp.GetParmName(param_names[p_idx]):
                LE_flags.append(vsp.GetParmVal(param_names[p_idx]))
            if 'UStart' == vsp.GetParmName(param_names[p_idx]):
                U_starts.append(vsp.GetParmVal(param_names[p_idx]))
            if 'UEnd' == vsp.GetParmName(param_names[p_idx]):
                U_ends.append(vsp.GetParmVal(param_names[p_idx]))
            if 'Length_C_Start' == vsp.GetParmName(param_names[p_idx]):
                chord_fractions.append(vsp.GetParmVal(param_names[p_idx]))

    # assign control surface parameters to wings. Outer most control surface on main/horizontal wing is assigned a aileron
    for cs_idx in range(num_cs):
        aileron_present = False
        if num_cs > 1:
            aileron_loc = np.argmax(np.array(U_starts))
            if cs_idx == aileron_loc:
                aileron_present = True
        if LE_flags[cs_idx] == 1.0:
            CS = RCAIDE.Library.Components.Wings.Control_Surfaces.Slat()
        else:
            if wing.vertical == True:
                CS = RCAIDE.Library.Components.Wings.Control_Surfaces.Rudder()
            else:
                if aileron_present:
                    CS = RCAIDE.Library.Components.Wings.Control_Surfaces.Aileron()
                else:
                    CS = RCAIDE.Library.Components.Wings.Control_Surfaces.Flap()
        CS.tag                 = tags[cs_idx]
        
        # Do some math to get U into a span fraction
        U_scale = 1/(N+2) # U scaling
        
        # Determine which segment the control surfaces begin and end, use floor
        segment_start = int(np.floor(U_starts[cs_idx]/U_scale)) - 1
        segment_end   = int(np.floor(U_ends[cs_idx]/U_scale)) - 1
        
        segment_normalized_start = U_starts[cs_idx]/U_scale - (segment_start+1)
        segment_normalized_end   = U_ends[cs_idx]/U_scale   - (segment_end+1)
        
        # calculate segment spans
        start_span   = breaks[segment_start+1] - breaks[segment_start]
        end_span     = breaks[segment_end+1]   - breaks[segment_end]
        
        # Calculate the offsets
        start_offset = breaks[segment_start]
        end_offset   = breaks[segment_end]
        
        # Calculate the end points
        span_start = segment_normalized_start * start_span + start_offset
        span_end   = segment_normalized_end   * end_span   + end_offset

        CS.span_fraction_start = np.max([span_start, 0])
        CS.span_fraction_end   = np.min([span_end,1])
        if CS.span_fraction_start > 1 or CS.span_fraction_end < 0:
            raise AssertionError("RCAIDE import of VSP files does not allow control surfaces defined for the wing caps.")

        CS.chord_fraction      = chord_fractions[cs_idx]
        CS.span                = (CS.span_fraction_end - CS.span_fraction_start)*wing.spans.projected
        wing.append_control_surface(CS)

    return wing


# ----------------------------------------------------------------------------------------------------------------------  
# write_vsp_wing
# ----------------------------------------------------------------------------------------------------------------------  
def write_vsp_wing(vehicle,wing, area_tags, fuel_tank_set_ind, OML_set_ind):
    """This write a given wing into OpenVSP format

    Assumptions:
    If wing segments are defined, they must cover the full span.
    (may work in some other cases, but functionality will not be maintained)

    Source:
    N/A

    Inputs:
    vehicle                                   [-] vehicle data structure
    wing.
      origin                                  [m] in all three dimensions
      spans.projected                         [m]
      chords.root                             [m]
      chords.tip                              [m]
      sweeps.quarter_chord                    [radians]
      twists.root                             [radians]
      twists.tip                              [radians]
      thickness_to_chord                      [-]
      dihedral                                [radians]
      tag                                     <string>
      Segments.*. (optional)
        twist                                 [radians]
        percent_span_location                 [-]  .1 is 10%
        root_chord_percent                    [-]  .1 is 10%
        dihedral_outboard                     [radians]
        sweeps.quarter_chord                  [radians]
        thickness_to_chord                    [-]
    area_tags                                 <dict> used to keep track of all tags needed in wetted area computation
    fuel_tank_set_index                       <int> OpenVSP object set containing the fuel tanks

    Outputs:
    area_tags                                 <dict> used to keep track of all tags needed in wetted area computation
    wing_id                                   <str>  OpenVSP ID for given wing

    Properties Used:
    N/A
    """
    wing_x = wing.origin[0][0]
    wing_y = wing.origin[0][1]
    wing_z = wing.origin[0][2]
    if wing.symmetric == True:
        span   = wing.spans.projected/2. # span of one side
    else:
        span   = wing.spans.projected
    root_chord = wing.chords.root
    tip_chord  = wing.chords.tip
    sweep      = wing.sweeps.quarter_chord / Units.deg
    sweep_loc  = 0.25
    root_twist = wing.twists.root / Units.deg
    tip_twist  = wing.twists.tip  / Units.deg
    root_tc    = wing.thickness_to_chord
    tip_tc     = wing.thickness_to_chord
    dihedral   = wing.dihedral / Units.deg

    # Check to see if segments are defined. Get count
    if len(wing.segments.keys())>0:
        n_segments = len(wing.segments.keys())
    else:
        n_segments = 0

    # Create the wing
    wing_id = vsp.AddGeom( "WING" )
    vsp.SetGeomName(wing_id, wing.tag)
    area_tags[wing.tag] = ['wings',wing.tag]

    # Make names for each section and insert them into the wing if necessary
    x_secs       = []
    x_sec_curves = []
    # n_segments + 2 will create an extra segment if the root segment is
    # included in the list of segments. This is not used and the tag is
    # removed when the segments are checked for this case.
    for i_segs in range(0,n_segments+2):
        x_secs.append('XSec_' + str(i_segs))
        x_sec_curves.append('XSecCurve_' + str(i_segs))

    # Apply the basic characteristics of the wing to root and tip
    if wing.symmetric == False:
        vsp.SetParmVal( wing_id,'Sym_Planar_Flag','Sym',0)
    if wing.vertical == True:
        vsp.SetParmVal( wing_id,'X_Rel_Rotation','XForm',90)
        dihedral = -dihedral # check for vertical tail, direction reverses from RCAIDE/AVL

    vsp.SetParmVal( wing_id,'X_Rel_Location','XForm',wing_x)
    vsp.SetParmVal( wing_id,'Y_Rel_Location','XForm',wing_y)
    vsp.SetParmVal( wing_id,'Z_Rel_Location','XForm',wing_z)

    # This ensures that the other VSP parameters are driven properly
    vsp.SetDriverGroup( wing_id, 1, vsp.SPAN_WSECT_DRIVER, vsp.ROOTC_WSECT_DRIVER, vsp.TIPC_WSECT_DRIVER )

    # Root chord
    vsp.SetParmVal( wing_id,'Root_Chord',x_secs[1],root_chord)

    # Sweep of the first section
    vsp.SetParmVal( wing_id,'Sweep',x_secs[1],sweep)
    vsp.SetParmVal( wing_id,'Sweep_Location',x_secs[1],sweep_loc)

    # Twists
    if n_segments != 0: 
        segment_keys   = list(wing.segments.keys())        
        if np.isclose(wing.segments[segment_keys[0]].percent_span_location,0.):
            vsp.SetParmVal( wing_id,'Twist',x_secs[0],wing.segments[segment_keys[0]].twist / Units.deg) # root
        else:
            vsp.SetParmVal( wing_id,'Twist',x_secs[0],root_twist) # root
        # The tips should write themselves
    else:
        vsp.SetParmVal( wing_id,'Twist',x_secs[0],root_twist) # root
        vsp.SetParmVal( wing_id,'Twist',x_secs[1],tip_twist) # tip


    # Figure out if there is an airfoil provided

    # Airfoils should be in Lednicer format
    # i.e. :
    #
    #EXAMPLE AIRFOIL
    # 3. 3.
    #
    # 0.0 0.0
    # 0.5 0.1
    # 1.0 0.0
    #
    # 0.0 0.0
    # 0.5 -0.1
    # 1.0 0.0

    # Note this will fail silently if airfoil is not in correct format
    # check geometry output

    airfoil_vsp_types = []
    if n_segments > 0: 
        for i in range(n_segments):
            if wing.segments[segment_keys[i]].airfoil !=  None:
                if type(wing.segments[segment_keys[i]].airfoil) == RCAIDE.Library.Components.Airfoils.Biconvex_Airfoil:
                    airfoil_vsp_types.append(vsp.XS_BICONVEX)
                else:
                    airfoil_vsp_types.append(vsp.XS_FILE_AIRFOIL)
            else:
                airfoil_vsp_types.append(vsp.XS_FILE_AIRFOIL)
    elif wing.airfoil !=  None:
        if type(wing.airfoil) == RCAIDE.Library.Components.Airfoils.Biconvex_Airfoil:
            airfoil_vsp_types.append(vsp.XS_BICONVEX)
        else:
            airfoil_vsp_types.append(vsp.XS_FILE_AIRFOIL)
    else:
        airfoil_vsp_types = [vsp.XS_FILE_AIRFOIL]

    if n_segments==0:
        if wing.airfoil != None:
            xsecsurf = vsp.GetXSecSurf(wing_id,0)
            vsp.ChangeXSecShape(xsecsurf,0,airfoil_vsp_types[0])
            vsp.ChangeXSecShape(xsecsurf,1,airfoil_vsp_types[0]) 
            xsec1 = vsp.GetXSec(xsecsurf,0)
            xsec2 = vsp.GetXSec(xsecsurf,1)
            vsp.ReadFileAirfoil(xsec1,wing.airfoil.coordinate_file)
            vsp.ReadFileAirfoil(xsec2,wing.airfoil.coordinate_file)
            vsp.Update()
    else:
        if wing.segments[segment_keys[0]].airfoil != None:
            xsecsurf = vsp.GetXSecSurf(wing_id,0)
            vsp.ChangeXSecShape(xsecsurf,0,airfoil_vsp_types[0])
            vsp.ChangeXSecShape(xsecsurf,1,airfoil_vsp_types[0]) 
            xsec1 = vsp.GetXSec(xsecsurf,0)
            xsec2 = vsp.GetXSec(xsecsurf,1)
            vsp.ReadFileAirfoil(xsec1,wing.segments[segment_keys[0]].airfoil.coordinate_file)
            vsp.ReadFileAirfoil(xsec2,wing.segments[segment_keys[0]].airfoil.coordinate_file)
            vsp.Update()

    # Thickness to chords
    vsp.SetParmVal( wing_id,'ThickChord','XSecCurve_0',root_tc)
    vsp.SetParmVal( wing_id,'ThickChord','XSecCurve_1',tip_tc)

    # Dihedral
    vsp.SetParmVal( wing_id,'Dihedral',x_secs[1],dihedral)

    # Span and tip of the section
    if n_segments>1:
        local_span    = span*wing.segments[segment_keys[-1]].percent_span_location
        sec_tip_chord = root_chord*wing.segments[segment_keys[-1]].root_chord_percent
        vsp.SetParmVal( wing_id,'Span',x_secs[1],local_span)
        vsp.SetParmVal( wing_id,'Tip_Chord',x_secs[1],sec_tip_chord)
    else:
        vsp.SetParmVal( wing_id,'Span',x_secs[1],span/np.cos(dihedral*Units.degrees))

    vsp.Update()

    if n_segments>0:
        if wing.segments[segment_keys[0]].percent_span_location==0.:
            x_secs[-1] = [] # remove extra section tag (for clarity)
            adjust = 0 # used for indexing
        else:
            adjust = 1
    else:
        adjust = 1


    # Loop for the number of segments left over
    for i_segs in range(1,n_segments+1):

        if (wing.segments[segment_keys[i_segs-1]] == wing.segments[segment_keys[-1]]) and (wing.segments[segment_keys[-1]].percent_span_location == 1.):
            break

        # Unpack
        dihedral_i = wing.segments[segment_keys[i_segs-1]].dihedral_outboard / Units.deg
        chord_i    = root_chord*wing.segments[segment_keys[i_segs-1]].root_chord_percent
        try:
            twist_i    = wing.segments[segment_keys[i_segs]].twist / Units.deg
            no_twist_flag = False
        except:
            no_twist_flag = True
        sweep_i    = wing.segments[segment_keys[i_segs-1]].sweeps.quarter_chord / Units.deg
        tc_i       = wing.segments[segment_keys[i_segs-1]].thickness_to_chord

        # Calculate the local span
        if i_segs == n_segments:
            span_i = span*(1 - wing.segments[segment_keys[i_segs-1]].percent_span_location)/np.cos(dihedral_i*Units.deg)
        else:
            span_i = span*(wing.segments[segment_keys[i_segs]].percent_span_location-wing.segments[segment_keys[i_segs-1]].percent_span_location)/np.cos(dihedral_i*Units.deg)

        # Insert the new wing section with specified airfoil if available
        if  wing.segments[segment_keys[i_segs-1]].airfoil != None:
            vsp.InsertXSec(wing_id,i_segs-1+adjust,airfoil_vsp_types[i_segs-1]) 
            xsecsurf = vsp.GetXSecSurf(wing_id,0)
            xsec = vsp.GetXSec(xsecsurf,i_segs+adjust)
            vsp.ReadFileAirfoil(xsec, wing.segments[segment_keys[i_segs-1]].airfoil.coordinate_file)
        else:
            vsp.InsertXSec(wing_id,i_segs-1+adjust,vsp.XS_FOUR_SERIES)

        # Set the parms


        # Find the id
        x_sec_id  = vsp.GetXSec(vsp.GetXSecSurf(wing_id, 0),i_segs+adjust)

        # Find the parm strings
        span_parm    = vsp.GetXSecParm(x_sec_id, 'Span')
        dih_parm     = vsp.GetXSecParm(x_sec_id, 'Dihedral')
        sweep_parm   = vsp.GetXSecParm(x_sec_id, 'Sweep')
        swp_loc_parm = vsp.GetXSecParm(x_sec_id, 'Sweep_Location')
        rt_ch_parm   = vsp.GetXSecParm(x_sec_id, 'Root_Chord')
        tc_parm      = vsp.GetXSecParm(x_sec_id, 'ThickChord')

        # Set the parm values
        vsp.SetParmVal(span_parm, span_i)
        vsp.SetParmVal(sweep_parm, sweep_i)
        vsp.SetParmVal(swp_loc_parm, sweep_loc)
        vsp.SetParmVal(rt_ch_parm, chord_i)
        vsp.SetParmVal(tc_parm, tc_i)
        
        # OpenVSP's rotation on vertical segmented wings is opposite of RCAIDE's
        if not wing.vertical:
            vsp.SetParmVal(dih_parm, dihedral_i)
        else:
            vsp.SetParmVal(dih_parm, -dihedral_i)

        if not no_twist_flag:
            twist_parm    = vsp.GetXSecParm(x_sec_id, 'Twist')
            vsp.SetParmVal(twist_parm,twist_i)

        if adjust and (i_segs == 1):
            vsp.Update()
            x_sec_id = vsp.GetXSec(vsp.GetXSecSurf(wing_id, 0),1)
            twist_parm    = vsp.GetXSecParm(x_sec_id, 'Twist')
            vsp.SetParmVal(twist_parm,wing.segments[segment_keys[i_segs-1]].twist / Units.deg)

        vsp.Update()

    if (n_segments != 0) and (wing.segments[segment_keys[-1]].percent_span_location == 1.):
        tip_chord = root_chord*wing.segments[segment_keys[-1]].root_chord_percent
        vsp.SetParmVal( wing_id,'Tip_Chord',x_secs[n_segments-1+adjust],tip_chord)
        vsp.SetParmVal( wing_id,'ThickChord',x_sec_curves[n_segments-1+adjust],wing.segments[segment_keys[-1]].thickness_to_chord)
        # twist is set in the normal loop
    else:
        vsp.SetParmVal( wing_id,'Tip_Chord',x_secs[-1-(1-adjust)],tip_chord)
        vsp.SetParmVal( wing_id,'Twist',x_secs[-1-(1-adjust)],tip_twist)
        # a single trapezoidal wing is assumed to have constant thickness to chord
    vsp.Update()
    vsp.SetParmVal(wing_id,'CapUMaxOption','EndCap',2.)
    vsp.SetParmVal(wing_id,'CapUMaxStrength','EndCap',1.)

    vsp.Update()

    if 'control_surfaces' in wing:
        for ctrl_surf in wing.control_surfaces:
            write_vsp_control_surface(wing,wing_id,ctrl_surf)


    if 'fuel_tanks' in wing:
        for tank in wing.fuel_tanks:
            write_wing_conformal_fuel_tank(vehicle,wing, wing_id, tank, fuel_tank_set_ind)

    vsp.SetSetFlag(wing_id, OML_set_ind, True)

    return area_tags, wing_id


# ----------------------------------------------------------------------------------------------------------------------  
# write_vsp_control_surface
# ----------------------------------------------------------------------------------------------------------------------  
def write_vsp_control_surface(wing,wing_id,ctrl_surf):
    """This writes a control surface in a wing.

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    wind_id              <str>
    ctrl_surf            [-]
    n_segments           int, number of wing segments

    Outputs:
    Operates on the active OpenVSP model, no direct output

    Properties Used:
    N/A
    """
    
    # determine the number of segments and where the breaks are
    if len(wing.segments.keys())>0:
        N = len(wing.segments.keys())-1
        
        # Do a for loop to find the location of breaks
        breaks = []
        for seg in wing.segments:
            breaks.append(seg.percent_span_location)
        
    else:
        N = 1
        # Location of breaks
        breaks = [0.,1.]
        
    breaks = np.array(breaks)    
    
    span_start = ctrl_surf.span_fraction_start
    span_end   = ctrl_surf.span_fraction_end
        

    cs_id =  vsp.AddSubSurf( wing_id, vsp.SS_CONTROL)
    param_names = vsp.GetSubSurfParmIDs(cs_id)
    for p_idx in range(len(param_names)):
        if 'LE_Flag' == vsp.GetParmName(param_names[p_idx]):
            if type(ctrl_surf) == RCAIDE.Library.Components.Wings.Control_Surfaces.Slat:
                vsp.SetParmVal(param_names[p_idx], 1.0)
            else:
                vsp.SetParmVal( param_names[p_idx], 0.0)
                
        # U values are only linear about the section in which they are defined
        
        # Identify the segments they start and end with, if there are no segments there is one segment
        segment_start = np.where(breaks<span_start)[0][-1]
        segment_end   = np.where(breaks>=span_end)[0][0] -1
        
        # Find where in the segment it starts and ends (normalized)
        start_span   = breaks[segment_start+1] - breaks[segment_start]
        end_span     = breaks[segment_end+1]   - breaks[segment_end]
        
        start_offset = breaks[segment_start]
        end_offset   = breaks[segment_end]
        
        segment_normalized_start = (span_start - start_offset)/start_span
        segment_normalized_end   = (span_end   - end_offset  )/end_span
        
        # Find the scaling for those segments
        U_scale = 1/(N + 2)

        # Used the normalized positions and the U_scaling to find the U location
        U_start = U_scale*segment_normalized_start + (segment_start + 1) * U_scale
        U_end   = U_scale*segment_normalized_end   + (segment_end   + 1) * U_scale
        
        # Voila!
        
        if 'UStart' == vsp.GetParmName(param_names[p_idx]):
            vsp.SetParmVal(param_names[p_idx], U_start)
        if 'UEnd' ==vsp.GetParmName(param_names[p_idx]):
            vsp.SetParmVal(param_names[p_idx], U_end)
        if 'Length_C_Start' == vsp.GetParmName(param_names[p_idx]):
            vsp.SetParmVal(param_names[p_idx], ctrl_surf.chord_fraction)
        if 'Length_C_End' == vsp.GetParmName(param_names[p_idx]):
            vsp.SetParmVal(param_names[p_idx], ctrl_surf.chord_fraction)
        if 'SE_Const_Flag' == vsp.GetParmName(param_names[p_idx]):
            vsp.SetParmVal(param_names[p_idx], 1.0)

    return

# ----------------------------------------------------------------------------------------------------------------------  
# write_wing_conformal_fuel_tank
# ----------------------------------------------------------------------------------------------------------------------  
def write_wing_conformal_fuel_tank(vehicle,wing, wing_id,fuel_tank,fuel_tank_set_ind):
    """This writes a conformal fuel tank in a wing.

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    vehicle                                     [-] vehicle data structure
    wing.segments.*.percent_span_location       [-]
    wing.spans.projected                        [m]
    wind_id                                     <str>
    fuel_tank.
      inward_offset                             [m]
      start_chord_percent                       [-] .1 is 10%
      end_chord_percent                         [-]
      start_span_percent                        [-]
      end_span_percent                          [-]
      fuel_type.density                         [kg/m^3]
    fuel_tank_set_ind                           <int>

    Outputs:
    Operates on the active OpenVSP model, no direct output

    Properties Used:
    N/A
    """
    # Unpack
    try:
        offset            = fuel_tank.inward_offset
        chord_trim_max    = 1.-fuel_tank.start_chord_percent
        chord_trim_min    = 1.-fuel_tank.end_chord_percent
        span_trim_max     = fuel_tank.end_span_percent
        span_trim_min     = fuel_tank.start_span_percent
        density           = fuel_tank.fuel_type.density
    except:
        print('Fuel tank does not contain parameters needed for OpenVSP geometry. Tag: '+fuel_tank.tag)
        return

    tank_id = vsp.AddGeom('CONFORMAL',wing_id)
    vsp.SetGeomName(tank_id, fuel_tank.tag)
    n_segments        = len(wing.segments.keys())
    if n_segments > 0.:
        seg_span_percents  = np.array([v['percent_span_location'] for (k,v)\
                                       in wing.segments.iteritems()])
        vsp_segment_breaks = np.linspace(0.,1.,n_segments)
    else:
        seg_span_percents = np.array([0.,1.])
    span = wing.spans.projected

    # Offset
    vsp.SetParmVal(tank_id,'Offset','Design',offset)

    for key, fuselage in vehicle.fuselages.items():
        width    = fuselage.width
        length   = fuselage.lengths.total
        hmax     = fuselage.heights.maximum
        height1  = fuselage.heights.at_quarter_length
        height2  = fuselage.heights.at_wing_root_quarter_chord
        height3  = fuselage.heights.at_three_quarters_length
        effdia   = fuselage.effective_diameter
        n_fine   = fuselage.fineness.nose
        t_fine   = fuselage.fineness.tail
        w_ac     = wing.aerodynamic_center

        w_origin = vehicle.wings.main_wing.origin
        w_c_4    = vehicle.wings.main_wing.chords.root/4.

        # Figure out the location x location of each section, 3 sections, end of nose, wing origin, and start of tail

        x1 = 0.25
        x2 = (w_origin[0]+w_c_4)/length
        x3 = 0.75

        fuse_id = vsp.AddGeom("FUSELAGE")
        vsp.SetGeomName(fuse_id, fuselage.tag)
        wing_id[fuselage.tag] = ['fuselages',fuselage.tag]

        # Set the origins:
        x = fuselage.origin[0][0]
        y = fuselage.origin[0][1]
        z = fuselage.origin[0][2]
        vsp.SetParmVal(fuse_id,'X_Location','XForm',x)
        vsp.SetParmVal(fuse_id,'Y_Location','XForm',y)
        vsp.SetParmVal(fuse_id,'Z_Location','XForm',z)
        vsp.SetParmVal(fuse_id,'Abs_Or_Relitive_flag','XForm',vsp.ABS) # misspelling from OpenVSP
        vsp.SetParmVal(fuse_id,'Origin','XForm',0.0)

    # Fuel tank chord bounds
    vsp.SetParmVal(tank_id,'ChordTrimFlag','Design',1.)
    vsp.SetParmVal(tank_id,'ChordTrimMax','Design',chord_trim_max)
    vsp.SetParmVal(tank_id,'ChordTrimMin','Design',chord_trim_min)

    # Fuel tank span bounds
    if n_segments>0:
        span_trim_max = get_vsp_trim_from_RCAIDE_trim(seg_span_percents,  vsp_segment_breaks,  span_trim_max)
        span_trim_min = get_vsp_trim_from_RCAIDE_trim(seg_span_percents,  vsp_segment_breaks,   span_trim_min)
    else:
        pass # no change to span_trim

    vsp.SetParmVal(tank_id,'UTrimFlag','Design',1.)
    vsp.SetParmVal(tank_id,'UTrimMax','Design',span_trim_max)
    vsp.SetParmVal(tank_id,'UTrimMin','Design',span_trim_min)

    # Set density
    vsp.SetParmVal(tank_id,'Density','Mass_Props',density)

    # Add to the full fuel tank set
    vsp.SetSetFlag(tank_id, fuel_tank_set_ind, True)

    return

# ---------------------------------------------------------------------------------------------------------------------- 
# get_vsp_trim_from_RCAIDE_trim
# ---------------------------------------------------------------------------------------------------------------------- 
def get_vsp_trim_from_RCAIDE_trim(seg_span_percents,vsp_segment_breaks,trim):
    """Compute OpenVSP span trim coordinates based on RCAIDE coordinates

    Assumptions:
    Wing does not have end caps

    Source:
    N/A

    Inputs:
    seg_span_percents   [-] range of 0 to 1
    vsp_segment_breaks  [-] range of 0 to 1
    trim                [-] range of 0 to 1 (RCAIDE value)

    Outputs:
    trim                [-] OpenVSP trim value

    Properties Used:
    N/A
    """
    # Determine max chord trim correction
    y_seg_ind                        = next(i for i,per_y in enumerate(seg_span_percents) if per_y > trim)
    segment_percent_of_total_span    = seg_span_percents[y_seg_ind] -  seg_span_percents[y_seg_ind-1]
    remaining_percent_within_segment = trim - seg_span_percents[y_seg_ind-1]
    percent_of_segment               = remaining_percent_within_segment/segment_percent_of_total_span
    trim                             = vsp_segment_breaks[y_seg_ind-1] +  (vsp_segment_breaks[y_seg_ind]-vsp_segment_breaks[y_seg_ind-1])*percent_of_segment
    return trim


# ---------------------------------------------------------------------------------------------------------------------- 
# convert_sweep
# ---------------------------------------------------------------------------------------------------------------------- 
def convert_sweep(sweep,sweep_loc,new_sweep_loc,AR,taper):
    """This converts arbitrary sweep into a desired sweep given
    wing geometry.

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    sweep               [degrees]
    sweep_loc           [unitless]
    new_sweep_loc       [unitless]
    AR                  [unitless]
    taper               [unitless]

    Outputs:
    quarter chord sweep

    Properties Used:
    N/A
    """
    sweep_LE  = np.arctan(np.tan(sweep)+4*sweep_loc* (1-taper)/(AR*(1+taper))) 
    new_sweep = np.arctan(np.tan(sweep_LE)-4*new_sweep_loc* (1-taper)/(AR*(1+taper)))

    return new_sweep



# RCAIDE/Library/Plots/Noise/plot_noise_level.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 

import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ---------------------------------------------------------------------------------------------------------------------- 
def plot_noise_level(noise_data,
                     noise_level = False,
                     save_figure = False,
                     save_filename="Sideline_Noise_Levels",
                     file_type=".png",
                     width = 11, height = 7): 
    """This plots the A-weighted Sound Pressure Level as a function of time at various aximuthal angles
    on the ground

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.
        frames.inertial.position_vector   - position vector of aircraft
        noise.
            total_SPL_dBA                 - total SPL (dbA)
            relative_microphone_locations - microphone locations

    Outputs:
    Plots

    Properties Used:
    N/A
    """      
    N_gm_y       = noise_data.microphone_y_resolution 
    gm           = noise_data.microphone_locations    
    gm_x         = gm[:,:,0]
    gm_y         = gm[:,:,1]    
    

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters) 
      
    fig         = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    axes        = fig.add_subplot(1,1,1) 
    
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,N_gm_y))  
      
    for k in range(N_gm_y):    
        axes.plot(gm_x[:,0]/Units.nmi, noise_level[:,k], marker = 'o', color = line_colors[k], label= r'mic at y = ' + str(round(gm_y[0,k],1)) + r' m' ) 
    axes.set_ylabel('SPL [dBA]')
    axes.set_xlabel('Range [nmi]')  
    set_axes(axes)
    axes.legend(loc='upper right')         
    if save_figure:
        plt.savefig(save_filename + ".png")    
        
    return fig


# RCAIDE/Library/Methods/Stability/Moment_of_Inertia/compute_aircraft_moment_of_inertia.py 
# 
# Created:  September 2024, A. Molloy

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia import compute_cuboid_moment_of_inertia, compute_cylinder_moment_of_inertia, compute_wing_moment_of_inertia

import RCAIDE
import numpy as  np 

# ------------------------------------------------------------------        
#  Component moments of inertia (MOI) tensors
# ------------------------------------------------------------------  
def compute_aircraft_moment_of_inertia(vehicle, CG_location, update_MOI=True): 
    ''' sums the moments of inertia of each component in the aircraft. Components summed: fuselages,
    wings (main, horizontal, tail + others), turbofan engines, batteries, motors, batteries, fuel tanks

    Assumptions:
    - All other components than those listed are insignificant

    Source:
 
    Inputs:
    - vehicle
    - Center of gravity

    Outputs:
    - Total aircraft moment of inertia tensor

    Properties Used:
    N/A
    '''    
    
    # ------------------------------------------------------------------        
    # Setup
    # ------------------------------------------------------------------      
    # Array to hold the entire aircraft's inertia tensor
    MOI_tensor = np.zeros((3, 3)) 
    MOI_mass = 0
    
    # ------------------------------------------------------------------        
    #  Fuselage(s)
    # ------------------------------------------------------------------      
    for fuselage in vehicle.fuselages:
        I, mass = fuselage.compute_moment_of_inertia(center_of_gravity = CG_location)
        MOI_tensor += I
        MOI_mass += mass
    
    # ------------------------------------------------------------------        
    #  Wing(s)
    # ------------------------------------------------------------------      
    for wing in vehicle.wings:
        I, mass = wing.compute_moment_of_inertia(mass=wing.mass_properties.mass, center_of_gravity =CG_location)
        MOI_tensor += I
        MOI_mass += mass
    
    # ------------------------------------------------------------------        
    #  Energy network
    # ------------------------------------------------------------------      
    I_network = np.zeros([3, 3]) 
    for network in vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor,RCAIDE.Library.Components.Propulsors.Electric_Rotor):
                motor   = propulsor.motor 
                I, mass = compute_cylinder_moment_of_inertia(motor.origin,motor.mass_properties.mass, 0, 0, 0,0, CG_location)
                I_network += I
                MOI_mass  += mass
                    
            if isinstance(propulsor,RCAIDE.Library.Components.Propulsors.Turbofan):
                I, mass= compute_cylinder_moment_of_inertia(propulsor.origin, propulsor.mass_properties.mass, propulsor.engine_length, propulsor.nacelle.diameter/2, 0, 0, CG_location)                    
                I_network += I
                MOI_mass += mass
            if isinstance(propulsor,RCAIDE.Library.Components.Propulsors.Turboprop):
                I, mass= compute_cylinder_moment_of_inertia(propulsor.origin, propulsor.mass_properties.mass, propulsor.engine_length, propulsor.engine_diameter/2, 0, 0, CG_location)                    
                I_network += I
                MOI_mass += mass
            if isinstance(propulsor,RCAIDE.Library.Components.Propulsors.ICE_Propeller):
                I, mass= compute_cylinder_moment_of_inertia(propulsor.origin, propulsor.mass_properties.mass, propulsor.engine_length, propulsor.engine_diameter/2, 0, 0, CG_location)                    
                I_network += I
                MOI_mass += mass
        
        for bus in network.busses: 
            for battery in bus.battery_modules: 
                I_battery, mass_battery = compute_cuboid_moment_of_inertia(battery.origin, battery.mass_properties.mass, battery.length, battery.width, battery.height, 0, 0, 0, CG_location)
                I_network += I_battery
                MOI_mass  += mass_battery         
                                 
        for fuel_line in network.fuel_lines:
            for fuel_tank in fuel_line.fuel_tanks:
                if isinstance(fuel_tank,RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks.Central_Fuel_Tank ): 
                    I, mass = compute_cuboid_moment_of_inertia(fuel_tank.origin, fuel_tank.fuel.mass_properties.mass, fuel_tank.length, fuel_tank.width, fuel_tank.height, 0, 0, 0, CG_location)
                    I_network += I
                    MOI_mass += mass
                if isinstance(fuel_tank,RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks.Wing_Fuel_Tank): 
                    I, mass =  compute_wing_moment_of_inertia(vehicle.wings["main_wing"], mass=fuel_tank.fuel.mass_properties.mass, center_of_gravity = CG_location, fuel_flag=True)
                    I_network += I
                    MOI_mass += mass                    
                else:
                    pass # TO DO
                        
    MOI_tensor += I_network    
    
    if update_MOI:
        vehicle.mass_properties.moments_of_inertia.tensor = MOI_tensor  
    return MOI_tensor,MOI_mass     


# RCAIDE/Library/Methods/Weights/Buildups/Common/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_boom_weight        import compute_boom_weight
from .compute_fuselage_weight    import compute_fuselage_weight
from .compute_rotor_weight       import compute_rotor_weight
from .compute_wing_weight        import compute_wing_weight
from .compute_wiring_weight      import compute_wiring_weight


# RCAIDE/Library/Components/Energy/Modulators/Fuel_Selector.py
#  
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# RCAIDE imports  
from RCAIDE.Library.Components import Component
 
# ----------------------------------------------------------------------------------------------------------------------
#  Fuel_Selector
# ----------------------------------------------------------------------------------------------------------------------  
class Fuel_Selector(Component):
    """
    Class for managing fuel flow control between tanks and engines
    
    Attributes
    ----------
    tag : str
        Identifier for the fuel selector (default: 'fuel_selector')
        
    efficiency : float
        Fuel transfer efficiency through the selector (default: 0.0)

    Notes
    -----
    The Fuel Selector controls fuel routing between multiple fuel tanks and engines,
    managing fuel distribution and tank selection during aircraft operation.

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks
        Fuel storage components
    RCAIDE.Library.Components.Energy.Distributors.Fuel_Line
        Fuel distribution components
    """
    
    def __defaults__(self):
        """
        Sets default values for fuel selector attributes
        
        Notes
        -----
        Initializes the selector with a default tag and zero efficiency. The efficiency
        should be set to an appropriate value based on the specific system configuration.
        """         

        self.tag              = 'fuel_selector'  
        self.efficiency       = 0.0       
     